#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2021 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import struct

from construct import ConstructError

from .filereceiver import FileReceiver
from ..telemetry import ax25


class FileReceiverMirSat1(FileReceiver):
    def file_id(self, chunk):
        return struct.unpack('<I', chunk[1:5])[0]

    def chunk_offset(self, chunk):
        return struct.unpack('<I', chunk[6:9] + b'\x00')[0]

    def chunk_data(self, chunk):
        return chunk[9:-2]

    def is_last_chunk(self, chunk):
        return chunk[0] & 0x20

    def parse_chunk(self, chunk):
        try:
            frame = ax25.parse(chunk)
        except ConstructError:
            return None
        if frame.header.pid == 0xBB:
            # TODO check CRC-16
            return frame.info
        return None


mirsat1 = FileReceiverMirSat1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2021 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, digital
import pmt

from ...hier.sync_to_pdu_packed import sync_to_pdu_packed
from ...hdlc_deframer import fcs_ok


# HDLC 0x7e flag
_syncword = '01111110'


class crop_and_check_crc(gr.basic_block):
    """
    Helper block to crop using the final 0x7e flag and check CRC-16
    """
    def __init__(self):
        gr.basic_block.__init__(
            self,
            name='crop_and_check_crc',
            in_sig=[],
            out_sig=[])
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('out'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        packet = pmt.u8vector_elements(msg)
        start = 0
        while True:
            try:
                idx = packet[start:].index(0x7e)
            except ValueError:
                return
            start += idx + 1
            p = packet[:idx]
            if fcs_ok(p):
                p = p[:-2]
                self.message_port_pub(
                    pmt.intern('out'),
                    pmt.cons(pmt.PMT_NIL, pmt.init_u8vector(len(p), p)))
                return


class yusat_deframer(gr.hier_block2):
    """
    Hierarchical block to deframe YUSAT ad-hoc AX.25-like protocol

    The input is a float stream of soft symbols. The output are PDUs
    with YUSAT frames.

    Args:
        options: Options from argparse
    """
    def __init__(self, options=None):
        gr.hier_block2.__init__(
            self,
            'yusat_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        self.message_port_register_hier_out('out')

        self.slicer = digital.binary_slicer_fb()
        # We hope that 256 bytes is long enough to contain the full packet
        self.deframer = sync_to_pdu_packed(
            packlen=256, sync=_syncword, threshold=0)
        self.crop = crop_and_check_crc()

        self.connect(self, self.slicer, self.deframer)
        self.msg_connect((self.deframer, 'out'), (self.crop, 'in'))
        self.msg_connect((self.crop, 'out'), (self, 'out'))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

'''
gr-satellites deframer components

The deframers transform soft symbols into frames, detecting packet
boundaries and performing error correction and checking as needed.

The input to these hierarchical blocks is a stream of soft symbols
and the output are PDUs with the frames.
'''

from .aalto1_deframer import aalto1_deframer
from .aausat4_deframer import aausat4_deframer
from .aistechsat_2_deframer import aistechsat_2_deframer
from .ao40_fec_deframer import ao40_fec_deframer
from .ao40_uncoded_deframer import ao40_uncoded_deframer
from .astrocast_fx25_deframer import astrocast_fx25_deframer
from .ax100_deframer import ax100_deframer
from .ax25_deframer import ax25_deframer
from .ax5043_deframer import ax5043_deframer
from .binar1_deframer import binar1_deframer
from .ccsds_concatenated_deframer import ccsds_concatenated_deframer
from .ccsds_rs_deframer import ccsds_rs_deframer
from .diy1_deframer import diy1_deframer
from .endurosat_deframer import endurosat_deframer
from .eseo_deframer import eseo_deframer
from .fossasat_deframer import fossasat_deframer
from .grizu263a_deframer import grizu263a_deframer
from .ideassat_deframer import ideassat_deframer
from .k2sat_deframer import k2sat_deframer
from .lilacsat_1_deframer import lilacsat_1_deframer
from .lucky7_deframer import lucky7_deframer
from .mobitex_deframer import mobitex_deframer
from .ngham_deframer import ngham_deframer
from .nusat_deframer import nusat_deframer
from .ops_sat_deframer import ops_sat_deframer
from .reaktor_hello_world_deframer import reaktor_hello_world_deframer
from .sat_3cat_1_deframer import sat_3cat_1_deframer
from .smogp_ra_deframer import smogp_ra_deframer
from .smogp_signalling_deframer import smogp_signalling_deframer
from .snet_deframer import snet_deframer
from .swiatowid_deframer import swiatowid_deframer
from .tt64_deframer import tt64_deframer
from .u482c_deframer import u482c_deframer
from .ua01_deframer import ua01_deframer
from .usp_deframer import usp_deframer
from .yusat_deframer import yusat_deframer



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, digital

from ... import decode_ra_code, check_tt64_crc
from ...hier.sync_to_pdu_soft import sync_to_pdu_soft
from ...utils.options_block import options_block


_syncword = '0010110111010100'
_syncword_new = '001011011101010001100011110001010011010110011001'

fec_sizes = {128: 260, 256: 514}


class smogp_ra_deframer(gr.hier_block2, options_block):
    """
    Hierarchical block to deframe the SMOG-P RA FEC frames

    The input is a float stream of soft symbols. The output are PDUs
    with frames.

    Args:
        frame_size: size of the frame before FEC (int)
        new_protocol: enable new protocol used in SMOG-1 (bool)
        syncword_threshold: number of bit errors allowed in syncword (int)
        options: Options from argparse
    """
    def __init__(self, frame_size, new_protocol=False,
                 syncword_threshold=None, options=None):
        gr.hier_block2.__init__(
            self,
            'smogp_ra_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        options_block.__init__(self, options)

        self.message_port_register_hier_out('out')

        if syncword_threshold is None:
            syncword_threshold = self.options.ra_syncword_threshold
        if syncword_threshold < 0:
            syncword_threshold = (self._default_sync_threshold_new
                                  if new_protocol
                                  else self._default_sync_threshold_old)
        syncword = _syncword_new if new_protocol else _syncword
        self.deframer = sync_to_pdu_soft(
            packlen=fec_sizes[frame_size] * 8, sync=syncword,
            threshold=syncword_threshold)
        self.fec = decode_ra_code(frame_size)
        if new_protocol:
            # CRC-16 ARC
            self.crc = check_tt64_crc(
                verbose=self.options.verbose_crc, packet_len=None, strip=False)

        self.connect(self, self.deframer)
        self.msg_connect((self.deframer, 'out'), (self.fec, 'in'))
        if new_protocol:
            self.msg_connect((self.fec, 'out'), (self.crc, 'in'))
            self.msg_connect((self.crc, 'ok'), (self, 'out'))
        else:
            self.msg_connect((self.fec, 'out'), (self, 'out'))

    _default_sync_threshold = -1
    _default_sync_threshold_old = 0
    _default_sync_threshold_new = 6

    @classmethod
    def add_options(cls, parser):
        """
        Adds SMOG-P RA deframer specific options to the argparse parser
        """
        parser.add_argument(
            '--ra_syncword_threshold', type=int,
            default=cls._default_sync_threshold,
            help='RA syncword bit errors [default=%(default)r]')
        parser.add_argument(
            '--verbose_crc', action='store_true', help='Verbose CRC decoder')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2018 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr
import numpy as np
import pmt

from .eseo_line_decoder import reflect_bytes as reflect


class reflect_bytes(gr.basic_block):
    """docstring for block reflect_bytes"""
    def __init__(self):
        gr.basic_block.__init__(
            self,
            name='reflect_bytes',
            in_sig=[],
            out_sig=[])
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('out'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        packet = np.array(pmt.u8vector_elements(msg), dtype='uint8')
        packet = np.unpackbits(packet)
        packet = reflect(packet)
        packet = np.packbits(packet)

        self.message_port_pub(
            pmt.intern('out'),
            pmt.cons(pmt.PMT_NIL, pmt.init_u8vector(len(packet), packet)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

'''
gr-satellites transport components

The transports transform PDUs with packets in a certain protocol
into PDUs with packets into another protocol. They are used to
implement upper layer network protocols.

The input to these hierarchical blocks are PDUs and the output
are PDUs as well.
'''

from .kiss_transport import kiss_transport



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 jgromes <gromes.jan@gmail.com>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *

from ..adapters import LinearAdapter


FuncId = Enum(
    Int8ul,
    RESP_PONG=0x20,
    RESP_REPEATED_MESSAGE=0x21,
    RESP_REPEATED_MESSAGE_CUSTOM=0x22,
    RESP_SYSTEM_INFO=0x23,
    RESP_PACKET_INFO=0x24,
    RESP_STATISTICS=0x25,
    RESP_FULL_SYSTEM_INFO=0x26,
    RESP_STORE_AND_FORWARD_ASSIGNED_SLOT=0x27,
    RESP_FORWARDED_MESSAGE=0x28,
    RESP_PUBLIC_PICTURE=0x29,
    RESP_DEPLOYMENT_STATE=0x2A,
    RESP_RECORDED_SOLAR_CELLS=0x2B,
    RESP_CAMERA_STATE=0x2C,
    RESP_RECORDED_IMU=0x2D,
    RESP_MANUAL_ACS_RESULT=0x2E,
    RESP_GPS_LOG=0x2F,
    RESP_GPS_LOG_STATE=0x30,
    RESP_FLASH_CONTENTS=0x31,
    RESP_CAMERA_PICTURE=0x32,
    RESP_CAMERA_PICTURE_LENGTH=0x33,
    RESP_GPS_COMMAND_RESPONSE=0x34,
    RESP_ACKNOWLEDGE=0x3F,
    )

Pong = Struct()

RepeatedMessage = Struct(
    'sender_id' / Int32ul,
    'msg' / HexDump(GreedyBytes),
    )

VoltageValue = LinearAdapter(1e3/20, Int8ul)
CurrentValue = LinearAdapter(1e6/10, Int16sl)
TempValue = LinearAdapter(1e3/10, Int16sl)

SystemInfo = Struct(
    'mppt_v_out' / VoltageValue,
    'mmpt_i_out' / CurrentValue,
    'onboard_time' / Int32ul,
    'power_cfg' / BitStruct(
        Padding(1),
        'mppt_keep_alive' / Flag,
        'mppt_temp_switch' / Flag,
        Padding(2),
        'lp_active' / Flag,
        'lp_enabled' / Flag,
        'tx_enabled' / Flag,
        ),
    'reset_ctr' / Int16ul,
    'xa_v_in' / VoltageValue,
    'xb_v_in' / VoltageValue,
    'za_v_in' / VoltageValue,
    'zb_v_in' / VoltageValue,
    'y_v_in' / VoltageValue,
    'batt_temp' / TempValue,
    'obc_temp' / TempValue,
    'flash_err_ctr' / Int32ul,
    )

SNRValue = LinearAdapter(1/4, Int8sl)
RSSIValue = LinearAdapter(1/-2, Int8ul)

PacketInfo = Struct(
    'last_snr' / SNRValue,
    'last_rssi' / RSSIValue,
    'lora_valid_ctr' / Int16ul,
    'lora_invalid_ctr' / Int16ul,
    'fsk_valid_ctr' / Int16ul,
    'fsk_invalid_ctr' / Int16ul,
    )

StatsEntryTemp = Struct(
    'min' / TempValue,
    'avg' / TempValue,
    'max' / TempValue,
    )

StatsEntryCurrent = Struct(
    'min' / CurrentValue,
    'avg' / CurrentValue,
    'max' / CurrentValue,
    )

StatsEntryVoltage = Struct(
    'min' / VoltageValue,
    'avg' / VoltageValue,
    'max' / VoltageValue,
    )

StatsEntryFloat = Struct(
    'min' / Float32l,
    'avg' / Float32l,
    'max' / Float32l,
    )

StatsTemperatures = Struct(
    'panel_y' / StatsEntryTemp,
    'top' / StatsEntryTemp,
    'bottom' / StatsEntryTemp,
    'battery' / StatsEntryTemp,
    'sec_battery' / StatsEntryTemp,
    )

StatsCurrents = Struct(
    'xa' / StatsEntryCurrent,
    'xb' / StatsEntryCurrent,
    'za' / StatsEntryCurrent,
    'zb' / StatsEntryCurrent,
    'y' / StatsEntryCurrent,
    'mppt' / StatsEntryCurrent,
    )

StatsVoltages = Struct(
    'xa' / StatsEntryVoltage,
    'xb' / StatsEntryVoltage,
    'za' / StatsEntryVoltage,
    'zb' / StatsEntryVoltage,
    'y' / StatsEntryVoltage,
    'mppt' / StatsEntryVoltage,
    )

StatsLight = Struct(
    'panel_y' / StatsEntryFloat,
    'top' / StatsEntryFloat,
    )

StatsAxes = Struct(
    'x' / StatsEntryFloat,
    'y' / StatsEntryFloat,
    'z' / StatsEntryFloat,
    )

StatsIMU = Struct(
    'gyro' / StatsAxes,
    'accel' / StatsAxes,
    'mag' / StatsAxes,
    )

Statistics = Struct(
    'flags' / BitStruct(
        Padding(3),
        'has_imu' / Flag,
        'has_light' / Flag,
        'has_volt' / Flag,
        'has_curr' / Flag,
        'has_temp' / Flag,
        ),
    'stats' / Struct(
        'temperature' / Optional(If(this._.flags.has_temp, StatsTemperatures)),
        'current' / Optional(If(this._.flags.has_curr, StatsCurrents)),
        'voltage' / Optional(If(this._.flags.has_volt, StatsVoltages)),
        'light' / Optional(If(this._.flags.has_light, StatsLight)),
        'imu' / Optional(If(this._.flags.has_imu, StatsIMU)),
        ),
    )

FullSystemInfo = Struct(
    'mppt_v_out' / VoltageValue,
    'mmpt_i_out' / CurrentValue,
    'onboard_time' / Int32ul,
    'power_cfg' / BitStruct(
        Padding(1),
        'mppt_keep_alive' / Flag,
        'mppt_temp_switch' / Flag,
        Padding(2),
        'lp_active' / Flag,
        'lp_enabled' / Flag,
        'tx_enabled' / Flag,
        ),
    'reset_ctr' / Int16ul,
    'xa_v_in' / VoltageValue,
    'xa_i_in' / CurrentValue,
    'xb_v_in' / VoltageValue,
    'xb_i_in' / CurrentValue,
    'za_v_in' / VoltageValue,
    'za_i_in' / CurrentValue,
    'zb_v_in' / VoltageValue,
    'zb_i_in' / CurrentValue,
    'y_v_in' / VoltageValue,
    'y_i_in' / CurrentValue,
    'y_temp' / TempValue,
    'obc_board_temp' / TempValue,
    'bottom_temp' / TempValue,
    'batt_temp' / TempValue,
    'sec_batt_temp' / TempValue,
    'mcu_temp' / TempValue,
    'y_lux' / Float32l,
    'top_lux' / Float32l,
    'x_bridge_fault' / Int8ul,
    'y_bridge_fault' / Int8ul,
    'z_bridge_fault' / Int8ul,
    'flash_err_ctr' / Int32ul,
    'fsk_rx_len' / Int8ul,
    'lora_rx_len' / Int8ul,
    'sensors' / BitStruct(
        'light_top' / Flag,
        'light_y' / Flag,
        'curr_mppt' / Flag,
        'curr_y' / Flag,
        'curr_zb' / Flag,
        'curr_za' / Flag,
        'curr_xb' / Flag,
        'curr_xa' / Flag,
        ),
    'last_adcs_res' / Int8ul,
    )

StoreAndForwardAssigned = Struct(
    'assigned_slot' / Int16ul,
    )

ForwardedMessage = Struct(
    'msg' / HexDump(GreedyBytes),
    )

DeploymentState = Struct(
    'deploy_ctr' / Int8ul,
    )

CameraState = Struct(
    'cam_state' / Int32ul,
    )

SampleAxes = Struct(
    'x' / Float32l,
    'y' / Float32l,
    'z' / Float32l,
    )

RecordedIMU = Struct(
    'flags' / BitStruct(
        Padding(5),
        'has_mag' / Flag,
        'has_accel' / Flag,
        'has_gyro' / Flag,
        ),
    'samples' / Struct(
        'gyro' / Optional(If(this._.flags.has_gyro, SampleAxes)),
        'accel' / Optional(If(this._.flags.has_accel, SampleAxes)),
        'mag' / Optional(If(this._.flags.has_mag, SampleAxes)),
        )
    )

ManualACSResult = Struct(
    'x_bridge_fault' / Int8ul,
    'y_bridge_fault' / Int8ul,
    'z_bridge_fault' / Int8ul,
    'adcs_elapsed' / Int32ul,
    )

GPSLog = Struct(
    'time' / Int32ul,
    'data' / HexDump(GreedyBytes),
    )

GPSLogState = Struct(
    'length' / Int32ul,
    'last_entry_addr' / Int32ul,
    'last_fix_addr' / Int32ul,
    )

FlashContents = Struct(
    'data' / HexDump(GreedyBytes),
    )

CameraPicture = Struct(
    'packet_id' / Int16ul,
    'data' / HexDump(GreedyBytes),
    )

CameraPictureLength = Struct(
    'len' / Int32ul,
    )

GPSCommandResponse = Struct(
    'resp' / HexDump(GreedyBytes),
    )

Acknowledge = Struct(
    'acked_func_id' / FuncId,
    'status' / Enum(
        Int8ul,
        ACK_OK=0x00,
        ACK_CALLSIGN_MISMATCH=0x01,
        ACK_RX_FAILED=0x02,
        ACK_FUNC_ID_FAILED=0x03,
        ACK_DECRYPT_FAILED=0x04,
        ACK_OPT_DATA_LEN_FAILED=0x05,
        ACK_FUNC_ID_UKNOWN=0x06,
        ),
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Athanasios Theocharis <athatheoc@gmail.com>
# This was made under ESA Summer of Code in Space 2019
# by Athanasios Theocharis, mentored by Daniel Estevez
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import numpy
from gnuradio import gr
import pmt
from . import telecommand
import array

class telecommand_primaryheader_adder(gr.basic_block):
    """
    docstring for block telecommand_primaryheader_adder
    """
    def __init__(self, bypass, control, spacecraft_id, virtual_channel_id):
        gr.basic_block.__init__(self,
            name="telecommand_primaryheader_adder",
            in_sig=[],
            out_sig=[])

        ##################################################
        # Parameters
        ##################################################
        self.transfer_frame_version = 0
        self.bypass = bypass
        self.control = control
        self.RSVD_spare = 0
        self.spacecraft_id = spacecraft_id
        self.virtual_channel_id = virtual_channel_id
        self.frame_length = 0
        self.frame_sequence_number = 0

        ##################################################
        # Blocks
        ##################################################
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('out'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print("[ERROR] Received invalid message type. Expected u8vector")
            return
        packet = pmt.u8vector_elements(msg)
        self.frame_length = len(packet) + 5
        if self.bypass == 1:
            self.frame_sequence_number = 0
        else:
            self.frame_sequence_number += 1

        finalHeader = array.array('B', telecommand.PrimaryHeader.build(dict(transfer_frame_version = self.transfer_frame_version,
                                                                            bypass = self.bypass,
                                                                            control = self.control,
                                                                            RSVD_spare = self.RSVD_spare,
                                                                            spacecraft_id = self.spacecraft_id,
                                                                            virtual_channel_id = self.virtual_channel_id,
                                                                            frame_length = self.frame_length,
                                                                            frame_sequence_number = self.frame_sequence_number))).tolist()

        finalPacket = numpy.append(finalHeader, packet)
        finalPacket = array.array('B', finalPacket[:])
        finalPacket = pmt.cons(pmt.PMT_NIL, pmt.init_u8vector(len(finalPacket), finalPacket))
        self.message_port_pub(pmt.intern('out'), finalPacket)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Athanasios Theocharis <athatheoc@gmail.com>
# This was made under ESA Summer of Code in Space 2019
# by Athanasios Theocharis, mentored by Daniel Estevez
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr
import pmt
from . import telemetry
from . import space_packet
import array

class telemetry_packet_reconstruction(gr.basic_block):
    """
    docstring for block telemetry_packet_reconstruction
    """
    def __init__(self):
        gr.basic_block.__init__(self,
            name="telemetry_packet_reconstruction",
            in_sig=[],
            out_sig=[])

        self.space_packet = []
        self.length_of_space_packet = 0
        ##################################################
        # Blocks
        ##################################################
        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('out'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print("[ERROR] Received invalid message type. Expected u8vector")
            return
        packet = bytearray(pmt.u8vector_elements(msg))
        size = len(packet) - 6
        try:
            header = telemetry.PrimaryHeader.parse(packet[:])
            if header.ocf_flag == 1:
                size -= 4
        except:
            print("Could not decode telemetry packet")
            return

        parsed = telemetry.FullPacket.parse(packet[:], size=size)

        payload = parsed.payload
        #The number 6 is used here, because that's the length of the Primary Header.
        #todo: Add a variable for this
        while len(payload) != 0:
            if len(self.space_packet) < 6:
                left = 6 - len(self.space_packet)
                self.space_packet.extend(payload[:left])
                payload = payload[left:]
            if len(self.space_packet) >= 6:
                self.length_of_space_packet = space_packet.PrimaryHeader.parse(bytearray(self.space_packet)).data_length
                left = self.length_of_space_packet + 6 - len(self.space_packet)
                self.space_packet.extend(payload[:left])
                payload = payload[left:]
                if 6 + self.length_of_space_packet == len(self.space_packet):
                    self.sendPacket()

    def sendPacket(self):
        packet = self.space_packet
        packet = array.array('B', packet[:])
        packet = pmt.cons(pmt.PMT_NIL, pmt.init_u8vector(len(packet), packet))
        self.message_port_pub(pmt.intern('out'), packet)
        self.length_of_space_packet = 0
        self.space_packet = []



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 jgromes <gromes.jan@gmail.com>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *

from ..adapters import LinearAdapter
from .fossasat import (
    VoltageValue, CurrentValue, TempValue, FuncId,
    StatsCurrents, StatsVoltages, StatsTemperatures, Pong, RepeatedMessage,
    SystemInfo, PacketInfo, Statistics, StoreAndForwardAssigned,
    ForwardedMessage, CameraPicture, DeploymentState)


SystemInfo = Struct(
    'v_batt' / VoltageValue,
    'i_chrg' / CurrentValue,
    'v_chrg' / VoltageValue,
    'uptime' / Int32ul,
    'power_cfg' / Struct(
        Padding(3),
        'tx_enabled' / Flag,
        'mppt_keep_alive' / Flag,
        'mppt_temp_switch' / Flag,
        'lp_enabled' / Flag,
        'lp_active' / Flag,
        ),
    'reset_ctr' / Int16ul,
    'v_panel_a' / VoltageValue,
    'v_panel_b' / VoltageValue,
    'v_panel_c' / VoltageValue,
    'batt_temp' / TempValue,
    'obc_temp' / TempValue,
    'mcu_temp' / Int8sl,
    )

Statistics = Struct(
    'flags' / BitStruct(
        'has_obc_temp' / Flag,
        'has_batt_temp' / Flag,
        'has_v_panel_c' / Flag,
        'has_v_panel_b' / Flag,
        'has_v_panel_a' / Flag,
        'has_v_batt' / Flag,
        'has_i_chrg' / Flag,
        'has_v_chrg' / Flag,
        ),
    'stats' / Struct(
        'v_chrg' / Optional(If(this._.flags.has_v_chrg, StatsVoltages)),
        'i_chrg' / Optional(If(this._.flags.has_i_chrg, StatsCurrents)),
        'v_batt' / Optional(If(this._.flags.has_v_batt, StatsVoltages)),
        'v_panel_a' / Optional(If(this._.flags.has_v_panel_a, StatsVoltages)),
        'v_panel_b' / Optional(If(this._.flags.has_v_panel_b, StatsVoltages)),
        'v_panel_c' / Optional(If(this._.flags.has_v_panel_c, StatsVoltages)),
        'batt_temp' / Optional(If(this._.flags.has_batt_temp,
                                  StatsTemperatures)),
        'obc_temp' / Optional(If(this._.flags.has_obc_temp,
                                 StatsTemperatures)),
        ),
    )

RecordedSolarCells = Struct(
    'samples' / GreedyRange(
        Struct(
            'v_panel_a' / VoltageValue,
            'v_panel_b' / VoltageValue,
            'v_panel_c' / VoltageValue,
            )
        )
    )

fossasat_1b = Struct(
    'callsign' / Const(b"FOSSASAT-1B"),
    'func_id' / FuncId,
    'opt_data_len' / Optional(Int8ul),
    'payload' / Switch(this.func_id, {
        'RESP_PONG': Pong,
        'RESP_REPEATED_MESSAGE': RepeatedMessage,
        'RESP_REPEATED_MESSAGE_CUSTOM': RepeatedMessage,
        'RESP_SYSTEM_INFO': SystemInfo,
        'RESP_PACKET_INFO': PacketInfo,
        'RESP_STATISTICS': Statistics,
        'RESP_STORE_AND_FORWARD_ASSIGNED_SLOT': StoreAndForwardAssigned,
        'RESP_FORWARDED_MESSAGE': ForwardedMessage,
        'RESP_PUBLIC_PICTURE': CameraPicture,
        'RESP_DEPLOYMENT_STATE': DeploymentState,
        'RESP_RECORDED_SOLAR_CELLS': RecordedSolarCells,
        }, default=HexDump(GreedyBytes)),
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2022 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

# This gives a wrapper around gnuradio.gr.types.*_t (GNU Radio 3.10)
# and gnuradio.blocks.*_t (GNU Radio 3.9), so that the same gr-satellites
# code base can work with both versions of GNU Radio

from gnuradio import gr

api_version = int(gr.api_version())
if api_version == 9:
    from gnuradio.blocks import byte_t, complex_t, float_t
elif api_version >= 10:
    from gnuradio.gr.gr_python.types import byte_t, complex_t, float_t
else:
    raise ValueError('unsupported GNU Radio API version', api_version)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import argparse
import functools
import itertools
import os
import shlex
import yaml

from gnuradio import gr, zeromq
import pmt

from ..components import datasinks
from ..components import datasources
from ..components import deframers
from ..components import demodulators
from ..components import transports
from ..satyaml import yamlfiles
from .. import pdu_add_meta


def set_options(cl, *args, **kwargs):
    """
    Given a class, returns a derived class with some fixed
    options set in the constructor.

    This is intended to generate GNU Radio blocks with some options set
    by deriving from blocks that allow for options.

    Args:
        cl: the base class to derive from (class)
        *args: arguments to pass to the __init__ method
        **kwargs: keyword arguments to pass to the __init__ method
    """
    class C(cl):
        __init__ = functools.partialmethod(cl.__init__, *args, **kwargs)

    return C


def try_add_options(x, parser):
    """
    Given an object x, calls x.add_options(parser) if add_options
    is an attribute of x.

    Args:
       x: an object
       parser: an argparser ArgumentParser
    """
    if hasattr(x, 'add_options'):
        x.add_options(parser)


def filter_translate_dict(d, key_translation):
    """
    Filter and translate the keys of a dictionary

    Args:
        d: a dictionary to filter and translate
        keys_translation: a dictionary of key translations
    """
    return {key_translation[k]: v
            for k, v in d.items() if k in key_translation}


class gr_satellites_flowgraph(gr.hier_block2):
    """
    gr-satellites decoder flowgraph

    Uses a YAML file with a satellite description to create a
    hierarchical flowgraph for that satellite. There are two modes of
    operation. If this is called from GRC, then only demodulation and
    deframing is done, getting all messages to the 'out' output port. If
    this is not called from GRC, then messages are routed to data sinks
    as appropriate.

    Args:
        file: filename of the YAML file to load (string)
        name: satellite name to search in all YAML files (string)
        norad: NORAD ID to search in all YAML files (int)
        samp_rate: sample rate (float)
        iq: use IQ or real input (bool)
        grc_block: whether this is called from GRC (bool)
        options: options from argparser
        config: configuration file from configparser
        pdu_in: use PDU input instead of samples (bool)
        dump_path: Path to dump internal signals to files (str)

    Note that exactly one of file, name and norad should be specified
    """
    def __init__(self, file=None, name=None, norad=None,
                 samp_rate=None, iq=False, grc_block=False,
                 options=None, config=None, pdu_in=False,
                 dump_path=None):
        gr.hier_block2.__init__(
            self,
            'gr_satellites_flowgraph',
            gr.io_signature(0, 0, 0)
            if pdu_in else
            gr.io_signature(1, 1,
                            gr.sizeof_gr_complex
                            if iq else gr.sizeof_float),
            gr.io_signature(0, 0, 0))

        self.samp_rate = samp_rate
        self.iq = iq
        self.grc_block = grc_block
        self.dump_path = dump_path
        self.config = config

        # Load up options, similarly to option block
        if type(options) is str:
            p = argparse.ArgumentParser(prog=self.__class__.__name__,
                                        conflict_handler='resolve')
            gr_satellites_flowgraph.add_options(p, file, name, norad)
            options = p.parse_args(shlex.split(options))

        self.options = options

        if pdu_in:
            self.message_port_register_hier_in('in')
        elif samp_rate is None:
            raise ValueError('samp_rate not specified')

        self.satyaml = satyaml = self.open_satyaml(file, name, norad)

        if grc_block:
            self.message_port_register_hier_out('out')
        else:
            self._datasinks = dict()
            self._additional_datasinks = list()
            do_telemetry = not (self.options is not None
                                and self.options.hexdump)
            for key, info in satyaml['data'].items():
                is_telemetry = ('telemetry' in info
                                or info == 'unknown')
                if not is_telemetry or do_telemetry:
                    self._init_datasink(key, info)
            self._init_additional_datasinks()
            self._transports = dict()
            if 'transports' in satyaml:
                for key, info in satyaml['transports'].items():
                    self._init_transport(key, info)

        if pdu_in:
            for sink in itertools.chain(
                    self._datasinks.values(), self._additional_datasinks):
                self.msg_connect((self, 'in'), (sink, 'in'))
        else:
            self._demodulators = dict()
            self._deframers = dict()
            self._taggers = dict()
            for key, transmitter in satyaml['transmitters'].items():
                self._init_demodulator_deframer(key, transmitter)

    def _init_datasink(self, key, info):
        """Initialize a datasink

        Initializes a datasink according to a SatYAML entry

        Args:
            key: the name of the datasink entry in SatYAML
            info: the body of the datasink entry in SatYAML
        """
        if 'decoder' in info:
            ds = getattr(datasinks, info['decoder'])
            try:
                datasink = ds(options=self.options)
            except TypeError:  # raised if ds doesn't have an options parameter
                datasink = ds()
        elif 'telemetry' in info:
            datasink = datasinks.telemetry_parser(info['telemetry'],
                                                  options=self.options)
        elif 'files' in info:
            datasink = datasinks.file_receiver(info['files'],
                                               options=self.options)
        elif 'image' in info:
            datasink = datasinks.file_receiver(
                info['image'], options=self.options, display=True)
        else:
            datasink = datasinks.hexdump_sink()
        self._datasinks[key] = datasink

    def _init_additional_datasinks(self):
        """Initialize additional datasinks

        Creates all the datasinks that are not explicitly indicated
        in the SatYAML (telemetry submit, KISS output, etc.)
        """
        if self.options is not None and self.options.kiss_out:
            self._additional_datasinks.append(
                datasinks.kiss_file_sink(self.options.kiss_out,
                                         bool(self.options.kiss_append),
                                         options=self.options))
        if self.options is not None and self.options.kiss_server:
            self._additional_datasinks.append(
                datasinks.kiss_server_sink(self.options.kiss_server_address,
                                           self.options.kiss_server))
        if self.options is not None and self.options.zmq_pub:
            self._additional_datasinks.append(
                zeromq.pub_msg_sink(self.options.zmq_pub))

        # The GR_SATELLITES_SUBMIT_TLM environment variable takes precendence
        # over the configuration to choose whether to enable telemetry
        # submission
        tlm_env = os.environ.get('GR_SATELLITES_SUBMIT_TLM')
        if tlm_env is not None:
            tlm_submit = bool(int(tlm_env))
        else:
            tlm_submit = self.config.getboolean('Groundstation', 'submit_tlm')
        if tlm_submit:
            self._additional_datasinks.extend(
                self.get_telemetry_submitters(self.satyaml, self.config,
                                              self.options))

        if self.options is not None and self.options.hexdump:
            self._additional_datasinks.append(datasinks.hexdump_sink())

    def _init_transport(self, key, info):
        """Initialize a datasink

        Initializes a transport according to a SatYAML entry and connects
        it to the appropriate datasink

        Args:
            key: the name of the transport entry in SatYAML
            info: the body of the transport entry in SatYAML
        """
        transport = self.get_transport(info['protocol'])()
        self._transports[key] = transport
        if not self.options.hexdump:
            for data in info['data']:
                self.msg_connect(
                    (transport, 'out'), (self._datasinks[data], 'in'))

    def _init_demodulator_deframer(self, key, transmitter):
        """Initialize a demodulator and deframer

        Creates a demodulator and deframer according to a SatYAML
        entry and connects the deframer to the data and transports

        Args:
            key: name of the transmitter entry in the SatYAML
            transmitter: transmitter entry in the SatYAML
        """
        baudrate = transmitter['baudrate']
        demod_options = ['deviation', 'af_carrier']
        demod_options = {k: k for k in demod_options}
        demodulator_additional_options = filter_translate_dict(transmitter,
                                                               demod_options)
        demodulator = self.get_demodulator(transmitter['modulation'])(
            baudrate=baudrate, samp_rate=self.samp_rate, iq=self.iq,
            dump_path=self.dump_path, options=self.options,
            **demodulator_additional_options)
        deframe_options = {
            'frame size': 'frame_size',
            'precoding': 'precoding',
            'RS basis': 'rs_basis',
            'RS interleaving': 'rs_interleaving',
            'convolutional': 'convolutional',
            'scrambler': 'scrambler',
            }
        deframer_additional_options = filter_translate_dict(transmitter,
                                                            deframe_options)
        deframer = self.get_deframer(transmitter['framing'])(
            options=self.options, **deframer_additional_options)
        self.connect(self, demodulator, deframer)
        self._demodulators[key] = demodulator
        self._deframers[key] = deframer

        self._connect_transmitter_to_data(key, transmitter, deframer)

    def _connect_transmitter_to_data(self, key, transmitter, deframer):
        """Connect a deframer to the datasinks and transports

        Connects a deframer to the datasinks and transports indicated in
        the SatYAML file

        Args:
            transmitter: the transmitter entry in SatYAML
            deframer: the deframer to connect
        """
        # Add a tagger
        meta = pmt.make_dict()
        meta = pmt.dict_add(meta, pmt.intern('transmitter'),
                            pmt.intern(key))
        tagger = pdu_add_meta(meta)
        self._taggers[key] = tagger
        self.msg_connect((deframer, 'out'), (tagger, 'in'))

        if self.grc_block:
            # If we are a GRC block we have no datasinks
            # so we connect directly to our output
            self.msg_connect((tagger, 'out'), (self, 'out'))
            return

        for s in self._additional_datasinks:
            self.msg_connect((tagger, 'out'), (s, 'in'))
        for data in transmitter.get('data', []):
            if data in self._datasinks:
                # The datasink may not exist if it's a telemetry parser
                # and we're running in hexdump mode
                self.msg_connect(
                    (tagger, 'out'), (self._datasinks[data], 'in'))
        for transport in transmitter.get('transports', []):
            self.msg_connect(
                (tagger, 'out'), (self._transports[transport], 'in'))

        if 'additional_data' in transmitter:
            for k, v in transmitter['additional_data'].items():
                # Add a tagger
                tagger = pdu_add_meta(meta)
                self._taggers[(key, k)] = tagger
                self.msg_connect((deframer, k), (tagger, 'in'))
                self.msg_connect((tagger, 'out'), (self._datasinks[v], 'in'))

    def get_telemetry_submitters(self, satyaml, config, options):
        """
        Returns a list of block instances of telemetry submitters

        The telemetry submitters are those appropriate for this satellite

        Args:
            satyaml: satellite YAML file, as returned by self.open_satyaml
            config: configuration file from configparser
        """
        norad = satyaml['norad']
        submitters = [
            datasinks.telemetry_submit(
                'SatNOGS', norad=norad, config=config, options=options)]
        for server in satyaml.get('telemetry_servers', []):
            port = None
            url = None
            if server.startswith('HIT '):
                port = server.split()[1]
                server = 'HIT'
            elif server.startswith('SIDS '):
                url = server.split()[1]
                server = 'SIDS'
            submitters.append(datasinks.telemetry_submit(
                server, norad=norad, port=port, url=url, config=config,
                options=options))
        return submitters

    def get_demodulator(self, modulation):
        return self._demodulator_hooks[modulation]

    def get_deframer(self, framing):
        return self._deframer_hooks[framing]

    def get_transport(self, protocol):
        return self._transport_hooks[protocol]

    @staticmethod
    def open_satyaml(file, name, norad):
        if sum([x is not None for x in [file, name, norad]]) != 1:
            raise ValueError(
                'exactly one of file, name and norad needs to be specified')

        if file is not None:
            satyaml = yamlfiles.get_yamldata(file)
        elif name is not None:
            satyaml = yamlfiles.search_name(name)
        else:
            satyaml = yamlfiles.search_norad(norad)

        return satyaml

    @classmethod
    def add_options(cls, parser, file=None, name=None, norad=None):
        satyaml = cls.open_satyaml(file, name, norad)

        demod_options = parser.add_argument_group('demodulation')
        deframe_options = parser.add_argument_group('deframing')
        data_options = parser.add_argument_group('data sink')

        for info in satyaml['data'].values():
            if 'decoder' in info:
                try_add_options(getattr(datasinks, info['decoder']),
                                data_options)
            if 'telemetry' in info:
                try_add_options(datasinks.telemetry_parser, data_options)
            if 'files' in info or 'image' in info:
                try_add_options(datasinks.file_receiver, data_options)

        for transmitter in satyaml['transmitters'].values():
            try_add_options(cls._demodulator_hooks[transmitter['modulation']],
                            demod_options)
            try_add_options(cls._deframer_hooks[transmitter['framing']],
                            deframe_options)

    # Default parameters
    _demodulator_hooks = {
        'AFSK': demodulators.afsk_demodulator,
        'FSK': demodulators.fsk_demodulator,
        'BPSK': demodulators.bpsk_demodulator,
        'BPSK Manchester': set_options(demodulators.bpsk_demodulator,
                                       manchester=True),
        'DBPSK': set_options(demodulators.bpsk_demodulator,
                             differential=True),
        'DBPSK Manchester': set_options(demodulators.bpsk_demodulator,
                                        differential=True, manchester=True),
        'FSK subaudio': set_options(demodulators.fsk_demodulator,
                                    subaudio=True),
        }
    _deframer_hooks = {
        'AX.25': set_options(deframers.ax25_deframer, g3ruh_scrambler=False),
        'AX.25 G3RUH': set_options(deframers.ax25_deframer,
                                   g3ruh_scrambler=True),
        'AX100 ASM+Golay': set_options(deframers.ax100_deframer, mode='ASM'),
        'AX100 Reed Solomon': set_options(deframers.ax100_deframer,
                                          mode='RS'),
        '3CAT-1': deframers.sat_3cat_1_deframer,
        'Astrocast FX.25 NRZ-I': set_options(deframers.astrocast_fx25_deframer,
                                             nrzi=True),
        'Astrocast FX.25 NRZ': set_options(deframers.astrocast_fx25_deframer,
                                           nrzi=False),
        'AO-40 FEC': deframers.ao40_fec_deframer,
        'AO-40 FEC short': set_options(deframers.ao40_fec_deframer,
                                       short_frames=True),
        'AO-40 FEC CRC-16-ARC': set_options(deframers.ao40_fec_deframer,
                                            crc=True),
        'AO-40 FEC CRC-16-ARC short': set_options(deframers.ao40_fec_deframer,
                                                  short_frames=True,
                                                  crc=True),
        'AO-40 uncoded': deframers.ao40_uncoded_deframer,
        'TT-64': deframers.tt64_deframer,
        'ESEO': deframers.eseo_deframer,
        'Lucky-7': deframers.lucky7_deframer,
        'Reaktor Hello World': deframers.reaktor_hello_world_deframer,
        'S-NET': deframers.snet_deframer,
        'SALSAT': set_options(deframers.snet_deframer, buggy_crc=False),
        'Swiatowid': deframers.swiatowid_deframer,
        'NuSat': deframers.nusat_deframer,
        'K2SAT': deframers.k2sat_deframer,
        'CCSDS Reed-Solomon': deframers.ccsds_rs_deframer,
        'CCSDS Concatenated': deframers.ccsds_concatenated_deframer,
        'CCSDS Uncoded': set_options(deframers.ccsds_rs_deframer, rs_en=False),
        'LilacSat-1': deframers.lilacsat_1_deframer,
        'AAUSAT-4': deframers.aausat4_deframer,
        'NGHam': set_options(deframers.ngham_deframer, decode_rs=True),
        'NGHam no Reed Solomon': set_options(deframers.ngham_deframer,
                                             decode_rs=False),
        'SMOG-P RA': deframers.smogp_ra_deframer,
        'SMOG-1 RA': set_options(deframers.smogp_ra_deframer,
                                 new_protocol=True),
        'SMOG-P Signalling': deframers.smogp_signalling_deframer,
        'SMOG-1 Signalling': set_options(deframers.smogp_signalling_deframer,
                                         new_protocol=True),
        'OPS-SAT': deframers.ops_sat_deframer,
        'U482C': deframers.u482c_deframer,
        'UA01': deframers.ua01_deframer,
        'Mobitex': deframers.mobitex_deframer,
        'Mobitex-NX': set_options(deframers.mobitex_deframer, nx=True),
        'FOSSASAT': deframers.fossasat_deframer,
        'AISTECHSAT-2': deframers.aistechsat_2_deframer,
        'AALTO-1': deframers.aalto1_deframer,
        'Grizu-263A': deframers.grizu263a_deframer,
        'IDEASSat': deframers.ideassat_deframer,
        'YUSAT': deframers.yusat_deframer,
        'AX5043': deframers.ax5043_deframer,
        'USP': deframers.usp_deframer,
        'DIY-1': deframers.diy1_deframer,
        'BINAR-1': deframers.binar1_deframer,
        'Endurosat': deframers.endurosat_deframer,
        }
    _transport_hooks = {
        'KISS': transports.kiss_transport,
        'KISS no control byte': set_options(transports.kiss_transport,
                                            control_byte=False),
        'KISS KS-1Q': set_options(transports.kiss_transport,
                                  control_byte=False, header_remove_bytes=3),
        }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import math

from construct import *

from .ax25 import Header as AX25Header
from ..adapters import LinearAdapter, AffineAdapter


Header = Struct(
    'command_id' / Int8ub,
    'seq_number' / Int8ub,
    'length' / Int8ub
    )

OperatingMode = Enum(Int8ub, off=0, test=1, await_launch=2, launch=3,
                     latency=4, initialization=5, commissioning=6, safe=7,
                     beacon=8, nominal=9, experiment=10)

BatteryWarning = Enum(BitsInteger(2), none=0, low=1, critical=2, high=3)


class TemperatureAdapter(Adapter):
    def __init__(self, *args, **kwargs):
        return Adapter.__init__(self, *args, **kwargs)

    def _decode(self, obj, context, path=None):
        return 0.336*(obj - 1708.1) if obj >= 1707.0 else \
          6.41 * (4.15 - math.sqrt(17.24 - 0.31 * (obj - 1712.2)))


Temperature = TemperatureAdapter(BitsInteger(12))


class BatteryTempAdapter(Adapter):
    def __init__(self, *args, **kwargs):
        return Adapter.__init__(self, *args, **kwargs)

    def _decode(self, obj, context, path=None):
        return 1.2 * (60 - math.sqrt(3600 - 1.72 * (2333 - obj)))


BatteryTemp = BatteryTempAdapter(BitsInteger(12))

PSUCurrent = AffineAdapter(232.6, 0.42, BitsInteger(12))

AnalogData = BitStruct(
    'BATT_TBAT_TM' / BatteryTemp[3],
    'reserved' / BitsInteger(12),
    'BATT_VBAT_TM' / AffineAdapter(264.1, -4039.2, BitsInteger(12)),
    'PSU_T_TM' / BitsInteger(12),
    'p3V3_TM' / BitsInteger(12),
    'p5V_TM' / BitsInteger(12),
    'p15V_TM' / BitsInteger(12),
    'n15V_TM' / BitsInteger(12),
    'PSU_Ip5V_TM' / PSUCurrent,
    'PSU_Ip15V_TM' / PSUCurrent,
    'PSU_In15V_TM' / PSUCurrent,
    'PSU_Ip3V3_TM' / PSUCurrent,
    'PDU_IVBUS_TM' / PSUCurrent,
    'PV_TPSXp_TM' / Temperature,
    'PV_TPSXn_TM' / Temperature,
    'PV_TPSYp_TM' / Temperature,
    'PV_TPSYn_TM' / Temperature,
    'PV_TPSZp_TM' / Temperature,
    'PV_ISPXp_TM' / AffineAdapter(810.64, 1688.3, BitsInteger(12)),
    'PV_ISPXn_TM' / AffineAdapter(656.02, 1622.3, BitsInteger(12)),
    'PV_ISPYp_TM' / AffineAdapter(853.8, 1798.4, BitsInteger(12)),
    'PV_ISPYn_TM' / AffineAdapter(810.64, 1688.3, BitsInteger(12)),
    'PV_ISPZp_TM' / AffineAdapter(638.81, 1571.8, BitsInteger(12)),
    'OBC_T_TM' / BitsInteger(12),
    'MGM_T_TM' / Temperature[3],
    'MGM_xyz_TM' / BitsInteger(12)[9],
    'MGM_TX_TM' / BitsInteger(12),
    'MODEM_T_TR_TM' / Temperature,
    'EBOX_T_INT_TM' / Temperature,
    'EBOX_T_EXT_TM' / Temperature,
    'BATT_T_EXT_TM' / Temperature,
    'BATT_T_INT_TM' / Temperature,
    'SS6_XYZ_TM' / AffineAdapter(17.7, -201.4, BitsInteger(12))[6],
    'RW_T_TM' / Temperature[2],
    'TP_TM' / Temperature[6],
    )

DigitalData = BitStruct(
    'battery_warning' / BatteryWarning,
    'DAS_p3V' / Flag,
    'DAS_p5V' / Flag,
    'DAS_p15V' / Flag,
    'DAS_n15V' / Flag,
    'DAS_p3V3' / Flag,
    'DAS_p5V_2' / Flag,
    'MGM1_p5V' / Flag,
    'MGM2_p5V' / Flag,
    'MGM3_p15V' / Flag,
    'MGM3_n15V' / Flag,
    'MGT_X_VBUS' / Flag,
    'TEMP_A_p5V' / Flag,
    'TEMP_B_p5V' / Flag,
    'MODEM_VBUS' / Flag,
    'RW_p5V' / Flag,
    'RW_VBUS' / Flag,
    'MTS_VBUS' / Flag,
    Padding(5)
    )

Time = LinearAdapter(4.0, Int32ub)

HousekeepingData = Struct(
    'operating_mode' / OperatingMode,
    'snapshot_time' / Time,
    'analog_data' / AnalogData,
    'digital_data' / DigitalData
    )

Data = Struct(
    'sent_time' / Time,
    'housekeeping' / HousekeepingData
    )

upmsat_2 = Struct(
    'ax25_header' / AX25Header,
    'header' / Header,
    'data' / Data
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from gnuradio import gr, blocks, digital, fec

from ... import decode_rs, aausat4_remove_fsm
from ...grpdu import pdu_to_tagged_stream, tagged_stream_to_pdu
from ...grtypes import byte_t
from ...hier.sync_to_pdu_soft import sync_to_pdu_soft
from ...hier.ccsds_descrambler import ccsds_descrambler
from ...utils.options_block import options_block


_syncword = '010011110101101000110100010000110101010101000010'


class aausat4_deframer(gr.hier_block2, options_block):
    """
    Hierarchical block to deframe AAUSAT-4 frames

    These frames use a terminated CCSDS convolutional code, a
    CCSDS scrambler and CCSDS Reed-Solomon.

    The input is a float stream of soft symbols. The output are PDUs
    with frames.

    Args:
        syncword_threshold: number of bit errors allowed in syncword (int)
        options: Options from argparse
    """
    def __init__(self, syncword_threshold=None, options=None):
        gr.hier_block2.__init__(
            self,
            'aausat4_deframer',
            gr.io_signature(1, 1, gr.sizeof_float),
            gr.io_signature(0, 0, 0))
        options_block.__init__(self, options)

        self.message_port_register_hier_out('out')

        if syncword_threshold is None:
            syncword_threshold = self.options.syncword_threshold

        self.deframer = sync_to_pdu_soft(
            packlen=1996 + 8, sync=_syncword, threshold=syncword_threshold)
        self.fsm = aausat4_remove_fsm()
        self.viterbi_short = fec.cc_decoder.make(
            1020, 7, 2, [79, -109], 0, -1, fec.CC_TERMINATED, False)
        self.viterbi_decoder_short = fec.async_decoder(
            self.viterbi_short, True, False, 1020//8)
        self.viterbi_long = fec.cc_decoder.make(
            1996, 7, 2, [79, -109], 0, -1, fec.CC_TERMINATED, False)
        self.viterbi_decoder_long = fec.async_decoder(
            self.viterbi_long, False, False, 1996//8)

        # Workaround for bug https://github.com/gnuradio/gnuradio/pull/2965
        # We use packed output for Viterbi short decoder and then
        # we unpack the PDUs.
        self.pdu2tag = pdu_to_tagged_stream(byte_t, 'packet_len')
        self.unpack = blocks.packed_to_unpacked_bb(1, gr.GR_MSB_FIRST)
        self.taglength = blocks.tagged_stream_multiply_length(
            gr.sizeof_char*1, 'packet_len', 8)
        self.tag2pdu = tagged_stream_to_pdu(byte_t, 'packet_len')

        self.scrambler = ccsds_descrambler()
        self.rs = decode_rs(False, 1)

        self.connect(self, self.deframer)
        self.msg_connect((self.deframer, 'out'), (self.fsm, 'in'))
        self.msg_connect((self.fsm, 'short'),
                         (self.viterbi_decoder_short, 'in'))
        self.msg_connect((self.fsm, 'long'), (self.viterbi_decoder_long, 'in'))

        # Workaround
        self.msg_connect((self.viterbi_decoder_short, 'out'),
                         (self.pdu2tag, 'pdus'))
        self.connect(self.pdu2tag, self.unpack, self.taglength, self.tag2pdu)
        self.msg_connect((self.tag2pdu, 'pdus'), (self.scrambler, 'in'))

        self.msg_connect((self.viterbi_decoder_long, 'out'),
                         (self.scrambler, 'in'))
        self.msg_connect((self.scrambler, 'out'), (self.rs, 'in'))
        self.msg_connect((self.rs, 'out'), (self, 'out'))

    _default_sync_threshold = 8

    @classmethod
    def add_options(cls, parser):
        """
        Adds AAUSAT-4 deframer specific options to the argparse parser
        """
        parser.add_argument(
            '--syncword_threshold', type=int,
            default=cls._default_sync_threshold,
            help='Syncword bit errors [default=%(default)r]')
        parser.add_argument(
            '--verbose_rs', action='store_true', help='Verbose RS decoder')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2021 The Regents of the University of Colorado
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from ..adapters import AffineAdapter, LinearAdapter
from construct import Adapter, BitsInteger, BitStruct, Container, Enum, Flag, \
                      Float32b, Int8ub, Int16ub, Int16sb, Int32ub, Padding, \
                      Struct

cute_pld_ccsds_header = BitStruct(
    'ccsds_version' / BitsInteger(3),
    'packet_type' / BitsInteger(1),
    'secondary_header_flag' / BitsInteger(1),
    'APID' / BitsInteger(11),
    'sequence_flags' / BitsInteger(2),
    'sequence_counter' / BitsInteger(14),
    'packet_length' / BitsInteger(16)
)

cute_pld_sw_stat = Struct(
    'ccsds_header' / cute_pld_ccsds_header,
    'shCoarse' / Int32ub,
    'shFine' / Int16ub,
    'pldSwVerMaj' / Int8ub,
    'pldSwVerMin' / Int8ub,
    'pldSwVerPatch' / Int8ub,
    'sdState' / BitStruct(
        Padding(6),
        'card1' / Enum(BitsInteger(1), DEAD=0, ALIVE=1),
        'card0' / Enum(BitsInteger(1), DEAD=0, ALIVE=1)
    ),
    'zynqTemp' / AffineAdapter(130.039011703511, 35520.15604681404, Int16ub),
    'zynqVccInt' / LinearAdapter(21739.1304347826, Int16ub),
    'zynqVccAux' / LinearAdapter(21739.1304347826, Int16ub),
    'zynqVccBram' / LinearAdapter(21739.1304347826, Int16ub),
    'zynqVccPint' / LinearAdapter(21739.1304347826, Int16ub),
    'zynqVccPAux' / LinearAdapter(21739.1304347826, Int16ub),
    'zynq_vccPdr0' / LinearAdapter(21739.1304347826, Int16ub),
    'zynqStatus' / Int8ub,
    'spare8' / Int8ub,
    'procTemp' / LinearAdapter(256.0, Int16sb),
    'ccdP5' / LinearAdapter(620.4551659097114, Int16ub),
    'ccdP15' / LinearAdapter(223.5870972357927, Int16ub),
    'ccdP32' / LinearAdapter(112.8099029045165, Int16ub),
    'ccdN5' / AffineAdapter(318.9131440052302, 2305.604898505892, Int16ub),
    'spare16' / Int16ub,
    'cmdRecvCount' / Int16ub,
    'cmdRjctCount' / Int16ub,
    'cmdSuccCount' / Int16ub,
    'cmdSuccOp' / Enum(Int16ub, NOOP=0, ARM=1, CMD_RST_STATS=2, CMD_XSUM=3,
                       CMD_ECHO_STATE=4, LOG_ROUTE=48, LOG_STATE=49,
                       LOG_RST_STATS=50, LOG_RESET_READ=52, LOG_ISSUE=53,
                       LOG_DUMP_INFO=54, LOG_SET_PUBLISH=55, MEM_DUMP=80,
                       MEM_LOAD=81, MEM_ERASE=82, MEM_XSUM=83, MEM_ABORT=84,
                       MEM_RESET=85, MEM_LOAD_DWORD=86, MEM_LOAD_WORD=87,
                       MEM_LOAD_BYTE=88, PKT_ISSUE=96, PKT_SET_RATE=97,
                       PKT_SET_STREAM=98, PKT_SET_PRIORITY=99,
                       PKT_QUERY_APID=100, PKT_SET_DELAY=101, TIME_RESET=112,
                       TBL_DUMP=128, TBL_LOAD=129, TBL_LOAD_START=130,
                       TBL_COMMIT=131, TBL_ABORT=132, TBL_VERIFY=133,
                       OS_RESET=144, OS_SEND_HK=145, FS_MKDIR=160,
                       FS_RMDIR=161, FS_RENAME=162, FS_DIRLIST=163,
                       FS_RM=164, FS_DUMP=165, FS_DEV_STAT=166, FS_LOAD=167,
                       FS_COMMIT=168, FS_FORMAT=169, FS_COPY=170,
                       FS_SD_STATE=171, TEC_ENA=176, TEC_DIS=177,
                       TEC_SET_SLEW_RATE=178, TEC_SET_TEMP=179,
                       TEC_READ_REG=180, TEC_WRITE_REG=181,
                       TEC_WRITE_DAC=182, TEC_SELECT_RTC=184, CCD_EXPOSE=192,
                       CCD_ABORT=193, CCD_SHUTTER_OVER=194,
                       CCD_SHUTTER_OPEN=195, CCD_SHUTTER_CLOSE=196,
                       CCD_SET_ID=197, CCD_COPY_IMG_SLOT=198,
                       CCD_COPY_REF_SLOT=199, CCD_ADD_REF=200,
                       CCD_MULT_REF=201, CCD_LOAD_REF=202, CCD_REPLAY=203,
                       CCD_DUMP=204, CCD_ENA=205, CCD_DIS=206, CCD_SET_DAC=207,
                       IMAGE_PROCESS=208, IMAGE_ABORT=209, MODE_STANDBY=224,
                       MODE_SCIENCE=225, CCD_STORE_REF=240),
    'cmdRjctOp' / Enum(Int16ub, NOOP=0, ARM=1, CMD_RST_STATS=2, CMD_XSUM=3,
                       CMD_ECHO_STATE=4, LOG_ROUTE=48, LOG_STATE=49,
                       LOG_RST_STATS=50, LOG_RESET_READ=52, LOG_ISSUE=53,
                       LOG_DUMP_INFO=54, LOG_SET_PUBLISH=55, MEM_DUMP=80,
                       MEM_LOAD=81, MEM_ERASE=82, MEM_XSUM=83, MEM_ABORT=84,
                       MEM_RESET=85, MEM_LOAD_DWORD=86, MEM_LOAD_WORD=87,
                       MEM_LOAD_BYTE=88, PKT_ISSUE=96, PKT_SET_RATE=97,
                       PKT_SET_STREAM=98, PKT_SET_PRIORITY=99,
                       PKT_QUERY_APID=100, PKT_SET_DELAY=101, TIME_RESET=112,
                       TBL_DUMP=128, TBL_LOAD=129, TBL_LOAD_START=130,
                       TBL_COMMIT=131, TBL_ABORT=132, TBL_VERIFY=133,
                       OS_RESET=144, OS_SEND_HK=145, FS_MKDIR=160,
                       FS_RMDIR=161, FS_RENAME=162, FS_DIRLIST=163,
                       FS_RM=164, FS_DUMP=165, FS_DEV_STAT=166, FS_LOAD=167,
                       FS_COMMIT=168, FS_FORMAT=169, FS_COPY=170,
                       FS_SD_STATE=171, TEC_ENA=176, TEC_DIS=177,
                       TEC_SET_SLEW_RATE=178, TEC_SET_TEMP=179,
                       TEC_READ_REG=180, TEC_WRITE_REG=181, TEC_WRITE_DAC=182,
                       TEC_SELECT_RTC=184, CCD_EXPOSE=192, CCD_ABORT=193,
                       CCD_SHUTTER_OVER=194, CCD_SHUTTER_OPEN=195,
                       CCD_SHUTTER_CLOSE=196, CCD_SET_ID=197,
                       CCD_COPY_IMG_SLOT=198, CCD_COPY_REF_SLOT=199,
                       CCD_ADD_REF=200, CCD_MULT_REF=201, CCD_LOAD_REF=202,
                       CCD_REPLAY=203, CCD_DUMP=204, CCD_ENA=205, CCD_DIS=206,
                       CCD_SET_DAC=207, IMAGE_PROCESS=208, IMAGE_ABORT=209,
                       MODE_STANDBY=224, MODE_SCIENCE=225, CCD_STORE_REF=240),
    'cmdFailCode' / Enum(Int8ub, SUCCESS=0, MODE=1, ARM=2, SOURCE=3, OPCODE=4,
                         METHOD=5, LENGTH=6, RANGE=7, CHECKSUM=8, PKT_TYPE=9),
    'armState' / BitStruct(
        Padding(6),
        'sc' / Enum(BitsInteger(1), OFF=0, ARMED=1),
        'dbg' / Enum(BitsInteger(1), OFF=0, ARMED=1)
    ),
    'logWriteCount' / Int16ub,
    'logDropCount' / Int16ub,
    'ccdEnaState' / Enum(Int8ub, DIS=0, ENA=1),
    'ccdCtrlState' / Enum(Int8ub, IDLE=0, ERASE=1, OPEN=2, EXPOSE=3, READOUT=4,
                          REPLAY=5, PROCESS=6, ENA=7, DIS=8),
    'ccdShutter' / Enum(Int8ub, CLOSED=0, OPEN=1),
    'shutterOverride' / Enum(Int8ub, DIS=0, ENA=1),
    'frameId' / Int32ub,
    'osCpuUsage' / Int16ub,
    'osCpuMax' / Int16ub,
    'timePPSCount' / Int16ub,
    'timeRecvCount' / Int16ub,
    'timeMissCount' / Int16ub,
    'fswMode' / Enum(Int8ub, STANDBY=0, SCIENCE=1),
    'tecState' / Enum(Int8ub, DIS=0, ENA=1),
    'tecSlewRate' / Float32b,
    'tecSetpoint' / Float32b,
    'tecCcdRtdTemp' / AffineAdapter(12.615295, 3276.7, Int16ub),
    'tecScRtd5Temp' / AffineAdapter(12.615295, 3276.7, Int16ub),
    'tecScRtd4Temp' / AffineAdapter(12.615295, 3276.7, Int16ub),
    'tecScRtd3Temp' / AffineAdapter(12.615295, 3276.7, Int16ub),
    'tecScRtd2Temp' / AffineAdapter(12.615295, 3276.7, Int16ub),
    'tecScRtd1Temp' / AffineAdapter(12.615295, 3276.7, Int16ub),
    'tecShutter' / LinearAdapter(13106.8, Int16ub),
    'tecVolt' / LinearAdapter(5234.636746976763, Int16ub),
    'tecAvgCurr' / Int16ub,
    'tecCurr' / Int16ub,
    'imgState' / Int8ub,
    'imgCurrProcType' / Enum(Int8ub, SKIP=0, RAW=1, BIN2D=2, BIAS=3, DARK=4,
                             OUTSPEC=5, SPEC1D=6, TRIM2D=7, XDISP=8)
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct.core import ConstructError

from .imagereceiver import ImageReceiver
from ..telemetry import by70_1 as tlm


class ImageReceiverBY701(ImageReceiver):
    def filename(self, fid):
        return f'{fid}.jpg'

    def parse_chunk(self, chunk):
        try:
            frame = tlm.parse(chunk)
        except ConstructError:
            return None
        return frame.camera


by70_1 = ImageReceiverBY701



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 Stefano Speretta <s.speretta@tudelft.nl>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *
from .ax25 import Header
from ..adapters import LinearAdapter
from ..adapters import AffineAdapter
import math


Address = Enum(BytesInteger(1),
               OBC=1,
               EPS=2,
               ADB=3,
               COMMS=4,
               ADCS=5,
               GROUND=8,
               LOBE=9,
               OBC2=10)

msgType = Enum(BytesInteger(1),
               Request=1,
               Reply=2)

service = Enum(BytesInteger(1),
               Beacon=80,
               Ping=17,
               Telemetry=3)

msgOutcome = Enum(BytesInteger(1),
                  OK=0,
                  Error=1)

EPSBusState = Enum(BytesInteger(1),
                   OFF=0,
                   ON=1)

sensorBitStatus = Enum(BytesInteger(1),
                       ERROR=0,
                       ACTIVE=1)

OBCStatus = BitStruct(
    'InternalState' / BitsInteger(4),
    'SoftwareImage' / BitsInteger(4)
    )

SubsystemStatus = BitStruct(
    Padding(4),
    'SoftwareImage' / BitsInteger(4)
    )

ResetCause = BitStruct(
    'SoftResetWDTTimerexpiration' / BitsInteger(1),
    'CPULockUp' / BitsInteger(1),
    'PORPowerSettle' / BitsInteger(1),
    'PORClockSettle' / BitsInteger(1),
    'VoltageAnomaly' / BitsInteger(1),
    'HardResetWDTWrongPassword' / BitsInteger(1),
    'HardResetWDTTimerexpiration' / BitsInteger(1),
    'SystemResetOutput' / BitsInteger(1),

    'SysCTLReboot' / BitsInteger(1),
    'NMIPin' / BitsInteger(1),
    'ExitLPM4.5' / BitsInteger(1),
    'ExitLPM3.5' / BitsInteger(1),
    'BadBandGapReference' / BitsInteger(1),
    'SupplySupervisorVccTrip' / BitsInteger(1),
    'VCCDetectorTrip' / BitsInteger(1),
    'SoftResetWDTWrongPassword' / BitsInteger(1),

    Padding(7),
    'DCOShortCircuitFault' / BitsInteger(1)
    )

# verify single bits
EPSSensorStatus = BitStruct(
    'BatteryINAStatus' / sensorBitStatus,
    'BatteryGGStatus' / sensorBitStatus,
    'InternalINAStatus' / sensorBitStatus,
    'UnregulatedINAStatus' / sensorBitStatus,
    'Bus1INAStatus' / sensorBitStatus,
    'Bus2INAStatus' / sensorBitStatus,
    'Bus3INAStatus' / sensorBitStatus,
    'Bus4INAStatus' / sensorBitStatus,
    'Bus4Error' / sensorBitStatus,
    'Bus3Error' / sensorBitStatus,
    'Bus2Error' / sensorBitStatus,
    'Bus1Error' / sensorBitStatus,
    'Bus4State' / EPSBusState,
    'Bus3State' / EPSBusState,
    'Bus2State' / EPSBusState,
    'Bus1State' / EPSBusState,
    'PanelYpINAStatus' / sensorBitStatus,
    'PanelYmINAStatus' / sensorBitStatus,
    'PanelXpINAStatus' / sensorBitStatus,
    'PanelXmINAStatus' / sensorBitStatus,
    'PanelYpTMPStatus' / sensorBitStatus,
    'PanelYmTMPStatus' / sensorBitStatus,
    'PanelXpTMPStatus' / sensorBitStatus,
    'PanelXmTMPStatus' / sensorBitStatus,
    'MpptYpINAStatus' / sensorBitStatus,
    'MpptYmINAStatus' / sensorBitStatus,
    'MpptXpINAStatus' / sensorBitStatus,
    'MpptXmINAStatus' / sensorBitStatus,
    'CellYpINAStatus' / sensorBitStatus,
    'CellYmINAStatus' / sensorBitStatus,
    'CellXpINAStatus' / sensorBitStatus,
    'CellXmINAStatus' / sensorBitStatus
    )

OBCSensorStatus = BitStruct(
    'INAStatus' / sensorBitStatus,
    'TMPStatus' / sensorBitStatus,
    Padding(6)
    )

ADBSensorStatus = BitStruct(
    'INAStatus' / sensorBitStatus,
    'TMPStatus' / sensorBitStatus,
    Padding(6)
    )

COMMSSensorStatus = BitStruct(
    'INAStatus' / sensorBitStatus,
    'TMPStatus' / sensorBitStatus,
    'TransmitINAStatus' / sensorBitStatus,
    'AmplifierINAStatus' / sensorBitStatus,
    'PhasingTMPStatus' / sensorBitStatus,
    'AmplifierTMPStatus' / sensorBitStatus,
    Padding(2)
    )

OBCTlmV2 = Struct(
    # Unit: degC
    'MCUTemperature' / LinearAdapter(10, BytesInteger(2, signed=True)),
    'SensorsStatus' / OBCSensorStatus,
    'BusVoltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    'BusCurrent' / LinearAdapter(1000, BytesInteger(2)),  # unit: A
    )

EPSTlmV2 = Struct(
    # Unit: degC
    'MCUTemperature' / LinearAdapter(10, BytesInteger(2, signed=True)),
    'Status' / EPSSensorStatus,
    # Unit: A
    'InternalINACurrent' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    'InternalINAVoltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    # Unit: A
    'UnregulatedINACurrent' / LinearAdapter(
        1000, BytesInteger(2, signed=True)),
    'UnregulatedINAVoltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    'BatteryGGVoltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    'BatteryINAVoltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    # Unit: A
    'BatteryINACurrent' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    'BatteryGGCapacity' / LinearAdapter(10, BytesInteger(2)),  # unit: mAh
    # Unit: degC
    'BatteryGGTemperature' / LinearAdapter(10, BytesInteger(2, signed=True)),
    # Unit: degC
    'BatteryTMP20Temperature' / LinearAdapter(
        10, BytesInteger(2, signed=True)),

    # Unit: A
    'Bus4Current' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    # Unit: A
    'Bus3Current' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    # Unit: A
    'Bus2Current' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    # Unit: A
    'Bus1Current' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    'Bus4Voltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    'Bus3Voltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    'Bus2Voltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    'Bus1Voltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V

    # Unit: A
    'PanelYpCurrent' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    # Unit: A
    'PanelYmCurrent' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    # Unit: A
    'PanelXpCurrent' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    # Unit: A
    'PanelXmCurrent' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    'PanelYpVoltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    'PanelYmVoltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    'PanelXpVoltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    'PanelXmVoltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    # Unit: degC
    'PanelYpTemperature' / LinearAdapter(10, BytesInteger(2, signed=True)),
    # Unit: degC
    'PanelYmTemperature' / LinearAdapter(10, BytesInteger(2, signed=True)),
    # Unit: degC
    'PanelXpTemperature' / LinearAdapter(10, BytesInteger(2, signed=True)),
    # Unit: degC
    'PanelXmTemperature' / LinearAdapter(10, BytesInteger(2, signed=True)),

    # Unit: A
    'MpptYpCurrent' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    # Unit: A
    'MpptYmCurrent' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    # Unit: A
    'MpptXpCurrent' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    # Unit: A
    'MpptXmCurrent' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    'MpptYpVoltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    'MpptYmVoltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    'MpptXpVoltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    'MpptXmVoltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V

    # Unit: A
    'CellYpCurrent' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    # Unit: A
    'CellYmCurrent' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    # Unit: A
    'CellXpCurrent' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    # Unit: A
    'CellXmCurrent' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    'CellYpVoltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    'CellYmVoltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    'CellXpVoltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    )

ADBTlmV2 = Struct(
    # Unit: degC
    'MCUTemperature' / LinearAdapter(10, BytesInteger(2, signed=True)),
    'SensorsStatus' / ADBSensorStatus,
    'Current' / LinearAdapter(1000, BytesInteger(2, signed=True)),  # unit: A
    'Voltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    # Unit: degC
    )

COMMSTlmV2 = Struct(
    # Unit: degC
    'MCUTemperature' / LinearAdapter(10, BytesInteger(2, signed=True)),
    'SensorsStatus' / COMMSSensorStatus,
    'Voltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    'Current' / LinearAdapter(1000, BytesInteger(2, signed=True)),  # unit: A
    # Unit: degC
    'Temperature' / LinearAdapter(10, BytesInteger(2, signed=True)),
    # Unit: dBm
    'ReceiverRSSI' / AffineAdapter(1, 21, BytesInteger(2, signed=True)),
    'TransmitVoltage' / LinearAdapter(1000, BytesInteger(2)),  # unit: V
    # Unit: A
    'TransmitCurrent' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    # Unit: V
    'AmplifierVoltage' / LinearAdapter(1000, BytesInteger(2)),
    # Unit: A
    'AmplifierCurrent' / LinearAdapter(1000, BytesInteger(2, signed=True)),
    # Unit: degC
    'PhasingBoardTemperature' / LinearAdapter(
        10, BytesInteger(2, signed=True)),
    )

OBCTlm = Struct(
    'Status' / OBCStatus,
    'BootCounter' / BytesInteger(1),
    'ResetCause' / ResetCause,
    'Uptime' / BytesInteger(4),  # unit: s
    'TotalUptime' / BytesInteger(4),  # unit: s
    'TLMVersion' / BytesInteger(1),
    'Telemetry' / Switch(this.TLMVersion, {
        2: OBCTlmV2})
    )

OBCTlmFull = Struct(
    'Status' / OBCStatus,
    'BootCounter' / BytesInteger(1),
    'ResetCause' / ResetCause,
    'Uptime' / BytesInteger(4),  # unit: s
    'TotalUptime' / BytesInteger(4),  # unit: s
    'TLMVersion' / BytesInteger(1),
    'Telemetry' / Switch(this.TLMVersion, {
        2: OBCTlmV2}),
    'BoardTemperature' / LinearAdapter(10, BytesInteger(2, signed=True))
    )

EPSTlm = Struct(
    'Status' / SubsystemStatus,
    'BootCounter' / BytesInteger(1),
    'ResetCause' / ResetCause,
    'Uptime' / BytesInteger(4),  # unit: s
    'TotalUptime' / BytesInteger(4),
    'TLMVersion' / BytesInteger(1),
    'Telemetry' / Switch(this.TLMVersion, {
        2: EPSTlmV2})
    )

EPSTlmFull = Struct(
    'Status' / SubsystemStatus,
    'BootCounter' / BytesInteger(1),
    'ResetCause' / ResetCause,
    'Uptime' / BytesInteger(4),  # unit: s
    'TotalUptime' / BytesInteger(4),
    'TLMVersion' / BytesInteger(1),
    'Telemetry' / Switch(this.TLMVersion, {
        2: EPSTlmV2}),
    'CellXmVoltage' / LinearAdapter(1000, BytesInteger(2))  # unit: V
    )

ADBTlm = Struct(
    'Status' / SubsystemStatus,
    'BootCounter' / BytesInteger(1),
    'ResetCause' / ResetCause,
    'Uptime' / BytesInteger(4),  # unit: s
    'TotalUptime' / BytesInteger(4),
    'TLMVersion' / BytesInteger(1),
    'Telemetry' / Switch(this.TLMVersion, {
        2: ADBTlmV2})
    )

ADBTlmFull = Struct(
    'Status' / SubsystemStatus,
    'BootCounter' / BytesInteger(1),
    'ResetCause' / ResetCause,
    'Uptime' / BytesInteger(4),  # unit: s
    'TotalUptime' / BytesInteger(4),
    'TLMVersion' / BytesInteger(1),
    'Telemetry' / Switch(this.TLMVersion, {
        2: ADBTlmV2}),
    # Unit: degC
    'Temperature' / LinearAdapter(10, BytesInteger(2, signed=True))
    )

COMMSTlm = Struct(
    'Status' / SubsystemStatus,
    'BootCounter' / BytesInteger(1),
    'ResetCause' / ResetCause,
    'Uptime' / BytesInteger(4),  # unit: s
    'TotalUptime' / BytesInteger(4),
    'TLMVersion' / BytesInteger(1),
    'Telemetry' / Switch(this.TLMVersion, {
        2: COMMSTlmV2})
    )

COMMSTlmFull = Struct(
    'Status' / SubsystemStatus,
    'BootCounter' / BytesInteger(1),
    'ResetCause' / ResetCause,
    'Uptime' / BytesInteger(4),  # unit: s
    'TotalUptime' / BytesInteger(4),
    'TLMVersion' / BytesInteger(1),
    'Telemetry' / Switch(this.TLMVersion, {
        2: COMMSTlmV2}),
    # Unit: degC
    'AmplifierTemperature' / LinearAdapter(10, BytesInteger(2, signed=True))
    )

PingMsg = Struct()

Beacon = Struct(
    'Destination' / Address,
    'Size' / BytesInteger(1),
    'BeaconSource' / Address,
    'Service' / service,
    'MessageType' / msgType,
    'MessageOutcome' / msgOutcome,
    'TLMSource' / Address,
    'TelemetryHeader' / Switch(this.TLMSource, {
        Address.OBC: OBCTlm,
        Address.EPS: EPSTlm,
        Address.ADB: ADBTlm,
        Address.COMMS: COMMSTlm})
    )

DirectCommand = Struct(
    'Service' / service,
    'MessageType' / msgType,
    'TelemetryHeader' / Switch(this.Service, {
        service.Ping: PingMsg,
        service.Telemetry: COMMSTlmFull})
    )

BusOverride = Struct(
    'CommandID' / Address,
    'Size' / BytesInteger(1),
    'MessageSource' / Address,
    'Service' / service,
    'TLMSource' / Address,
    'TelemetryHeader' / Switch(this.TLMSource, {
        Address.OBC: OBCTlmFull,
        Address.EPS: EPSTlmFull,
        Address.ADB: ADBTlmFull,
        Address.COMMS: COMMSTlm})
    )

BufferedCommand = Struct(
    'CommandID' / Address,
    'Destination' / Address,
    'Size' / BytesInteger(1),
    'Source' / Address,
    'Service' / service,
    'Request' / msgType,
    'TelemetryHeader' / Switch(this.Service, {
        Address.OBC: OBCTlmFull,
        Address.EPS: EPSTlmFull,
        Address.ADB: ADBTlmFull,
        Address.COMMS: COMMSTlmFull})
    )

delfipq = Struct(
    'ax25_header' / Header,
    'FrameID' / BytesInteger(1),
    'packet' / Switch(this.FrameID, {
        0: Beacon,
        1: DirectCommand,
        2: BusOverride,
        3: BufferedCommand})
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2018,2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import datetime

from construct import *
import construct

from ..adapters import LinearAdapter

# See
# https://www.raumfahrttechnik.tu-berlin.de/fileadmin/fg169/
# amateur-radio/TUBiX10-COM.pdf
# for documentation

LTUFrameHeader = BitStruct(
    'SrcId' / BitsInteger(7),
    'DstId' / BitsInteger(7),
    'FrCntTx' / BitsInteger(4),
    'FrCntRx' / BitsInteger(4),
    'SNR' / BitsInteger(4),
    'AiTypeSrc' / BitsInteger(4),
    'AiTypeDst' / BitsInteger(4),
    'DfcId' / BitsInteger(2),
    'Caller' / Flag,
    'Arq' / Flag,
    'PduTypeId' / Flag,
    'BchRq' / Flag,
    'Hailing' / Flag,
    'UdFl1' / Flag,
    'PduLength' / BitsInteger(10),
    'CRC13' / BitsInteger(13),
    'CRC5' / BitsInteger(5),
    Padding(2)
    )


class TimeAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        d = int((obj - datetime.datetime(2000, 1, 1))*2)
        return Container(
            days=d.days,
            milliseconds=d.seconds * 1000 + d.microseconds / 1000)

    def _decode(self, obj, context, path=None):
        return (datetime.datetime(2000, 1, 1)
                + datetime.timedelta(seconds=float(obj)/2.0))


TimeStamp = TimeAdapter(BitsInteger(32, swapped=True))


SNETFrameHeaderExtension = BitStruct(
    'VersNo' / BitsInteger(2),
    'DFCID' / BitsInteger(2),
    'ExtensionRFU' / BitsInteger(4),
    'ChannelInfo' / BitsInteger(8),
    'QoS' / Flag,
    'PDUTypeID' / Flag,
    'ARQ' / Flag,
    'ControlRFU' / BitsInteger(5),
    'TimeTagSub' / BitsInteger(16),
    'SCID' / BitsInteger(10),
    'SeqNo' / BitsInteger(14)
    )

SNETFrameHeader = BitStruct(
    Const(0b111100110101000000, BitsInteger(18)),
    'CRC' / BitsInteger(14),
    'FCIDMajor' / BitsInteger(6),
    'FCIDSub' / BitsInteger(10),
    'Urgent' / Flag,
    'Extended' / Flag,
    'CheckCRC' / Flag,
    'Multiframe' / Flag,
    'TimeTaggedSetting' / Flag,
    'TimeTagged' / Flag,
    'DataLength' / BitsInteger(10),
    'TimeTag' / If(lambda c: c.TimeTagged, TimeStamp),
    'Extension' / If(lambda c: c.Extended, SNETFrameHeaderExtension)
    )

Battery = Struct(
    'V_BAT' / LinearAdapter(2, Int16sl),
    'A_IN_CHARGER' / LinearAdapter(12, Int16sl),
    'A_OUT_CHARGER' / LinearAdapter(6, Int16sl)
    )

BatteryCurrents = Struct(
    'A_IN' / LinearAdapter(12, Int16sl),
    'A_OUT' / LinearAdapter(12, Int16sl)
    )

EPSTelemetry = Struct(
    'CUR_SOL' / LinearAdapter(50, Int16sl)[6],
    'V_SOL' / Int16sl,
    'BATTERIES' / Battery[2],
    'V_SUM' / LinearAdapter(2, Int16sl),
    'V_3V3' / LinearAdapter(8, Int16sl),
    'V_5V' / LinearAdapter(5, Int16sl),
    'TEMP_BATT' / LinearAdapter(256, Int16sl)[2],
    'TEMP_OBC' / Int16sl,
    'A_OBC' / Int16ul,
    'V_OBC' / Int16ul,
    'BATT_CURRENTS' / BatteryCurrents[2]
    )

ADCSFlags = BitStruct(
    Padding(4),
    'AttDetTrackIGRFDeltaB' / Flag,
    'AttDetSuseAlbedoTracking' / Flag,
    'SUSE1AlbedoFlag' / Flag,
    'SUSE2AlbedoFlag' / Flag,
    'SUSE3AlbedoFlag' / Flag,
    'SUSE4AlbedoFlag' / Flag,
    'SUSE5AlbedoFlag' / Flag,
    'SUSE6AlbedoFlag' / Flag,
    'AttDetAutoVirtualizeMFSA' / Flag,
    'AttDetAutoVirtualizeSUSEA' / Flag,
    'AttDetNarrowVectors' / Flag,
    'AttDetMismatchingVectors' / Flag
)

ADCSTelemetry = Struct(
    'iModeChkListThisStepActive' / Int8sl,
    'iAttDetFinalState' / Int8ul,
    'iSensorArrayAvailStatusGA' / Int8ul,
    'iSensorArrayAvailStatusMFSA' / Int8ul,
    'iSensorArrayAvailStatusSUSEA' / Int8ul,
    'iActArrayAvailStatusRWA' / Int8ul,
    'iActArrayAvailStatusMATA' / Int8ul,
    'AttDetMfsDistCorrMode' / Int8ul,
    'AttDetSuseDistCorrMode' / Int8ul,
    'flags' / ADCSFlags,
    'omegaXOptimal_SAT' / LinearAdapter(260, Int16sl),
    'omegaYOptimal_SAT' / LinearAdapter(260, Int16sl),
    'omegaZOptimal_SAT' / LinearAdapter(260, Int16sl),
    'magXOptimal_SAT' / LinearAdapter(0.1, Int16sl),
    'magYOptimal_SAT' / LinearAdapter(0.1, Int16sl),
    'magZOptimal_SAT' / LinearAdapter(0.1, Int16sl),
    'sunXOptimal_SAT' / LinearAdapter(32000, Int16sl),
    'sunYOptimal_SAT' / LinearAdapter(32000, Int16sl),
    'sunZOptimal_SAT' / LinearAdapter(32000, Int16sl),
    'dCtrlTorqueRWAx_SAT_lr' / LinearAdapter(38484, Int8ul),
    'dCtrlTorqueRWAy_SAT_lr' / LinearAdapter(38484, Int8ul),
    'dCtrlTorqueRWAz_SAT_lr' / LinearAdapter(38484, Int8ul),
    'dCtrlMagMomentMATAx_SAT_lr' / LinearAdapter(127, Int8ul),
    'dCtrlMagMomentMATAy_SAT_lr' / LinearAdapter(127, Int8ul),
    'dCtrlMagMomentMATAz_SAT_lr' / LinearAdapter(127, Int8ul),
    'iReadTorqueRWx_MFR' / LinearAdapter(9696969, Int16ul),
    'iReadTorqueRWy_MFR' / LinearAdapter(9696969, Int16ul),
    'iReadTorqueRWz_MFR' / LinearAdapter(9696969, Int16ul),
    'iReadRotSpeedRWx_MFR' / Int16ul,
    'iReadRotSpeedRWy_MFR' / Int16ul,
    'iReadRotSpeedRWz_MFR' / Int16ul,
    'SGP4LatXPEF' / LinearAdapter(355, Int16ul),
    'SGP4LongYPEF' / LinearAdapter(177, Int16ul),
    'SGP4AltPEF' / LinearAdapter(0.25, Int8ul),
    'AttitudeErrorAngle' / LinearAdapter(177, Int16ul),
    'TargetData_Distance' / Int16ul,
    'TargetData_ControlIsActive' / Int8ul  # flag, really
    )


snet = Struct(
    'header' / SNETFrameHeader,
    'telemetry' / Switch(lambda c: (c.header.FCIDMajor, c.header.FCIDSub), {
        (0, 0): ADCSTelemetry,
        (9, 0): EPSTelemetry,
    }, default=Pass))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017,2020 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *

SatID = Enum(
    BitsInteger(2),
    FC1EM=0,
    FC2=1,
    FC1FM=2,
    extended=3,
    )

FrameType = Enum(
    BitsInteger(6),
    WO1=0,
    WO2=1,
    WO3=2,
    WO4=3,
    WO5=4,
    WO6=5,
    WO7=6,
    WO8=7,
    WO9=8,
    WO10=9,
    WO11=10,
    WO12=11,
    HR1=12,
    FM1=13,
    FM2=14,
    FM3=15,
    HR2=16,
    FM4=17,
    FM5=18,
    FM6=19,
    HR3=20,
    FM7=21,
    FM8=22,
    FM9=23)

FrameTypeNayif1 = Enum(
    BitsInteger(6),
    WO1=0,
    WO2=1,
    WO3=2,
    WO4=3,
    WO5=4,
    WO6=5,
    WO7=6,
    WO8=7,
    WO9=8,
    WO10=9,
    WO11=10,
    WO12=11,
    HR1=12,
    HR2=13,
    HR3=14,
    HR4=15,
    HR5=16,
    FM1=17,
    FM2=18,
    FM3=19,
    FM4=20,
    FM5=21,
    FM6=22,
    FM7=23)


class FrameTypeAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return obj.value

    def _decode(self, obj, context, path=None):
        return FrameType(obj)


FrameTypeField = FrameTypeAdapter(BitsInteger(6))


Header = BitStruct(
    'satid' / SatID,
    'frametype' / IfThenElse(lambda c: c.satid != 'extended',
                             FrameType, FrameTypeNayif1),
    )

EPSFC1 = Struct(
    'photovoltage' / BitsInteger(16)[3],
    'photocurrent' / BitsInteger(16),
    'batteryvoltage' / BitsInteger(16),
    'systemcurrent' / BitsInteger(16),
    'rebootcount' / BitsInteger(16),
    'softwareerrors' / BitsInteger(16),
    'boostconvertertemp' / Octet[3],
    'batterytemp' / Octet,
    'latchupcount5v' / Octet,
    'latchupcount3v3' / Octet,
    'resetcause' / Octet,
    'MPPTmode' / Octet,
    )

EPSNayif1 = Struct(
    'photovoltage' / BitsInteger(14)[3],
    'batteryvoltage' / BitsInteger(14),
    'photocurrent' / BitsInteger(10)[3],
    'totalphotocurrent' / BitsInteger(10),
    'systemcurrent' / BitsInteger(10),
    'rebootcount' / Octet,
    'boostconvertertemp' / Octet[3],
    'batterytemp' / Octet,
    'latchupcount5v' / Octet,
    'channelcurrent5v' / Octet,
    'resetcause' / BitsInteger(4),
    'MPTTmode' / BitsInteger(4),
    )

iMTQMode = Enum(
    BitsInteger(2),
    idle=0,
    selftest=1,
    detumble=2,
    )

iMTQ = Struct(
    'imtqmode' / iMTQMode,
    'imtqerrorcode' / BitsInteger(3),
    'imtqconfigurationset' / BitsInteger(1),
    'imtqmcutemperature' / Octet,
    )


class TempXpAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int((obj-158.239)/-0.2073)

    def _decode(self, obj, context, path=None):
        return -0.2073*obj + 158.239


TempXp = TempXpAdapter(BitsInteger(10))


class TempXmAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int((obj-159.227)/-0.2083)

    def _decode(self, obj, context, path=None):
        return -0.2083*obj + 159.227


TempXm = TempXmAdapter(BitsInteger(10))


class TempYpAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int((obj-158.656)/-0.2076)

    def _decode(self, obj, context, path=None):
        return -0.2076*obj + 158.656


TempYp = TempYpAdapter(BitsInteger(10))


class TempYmAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int((obj-159.045)/-0.2087)

    def _decode(self, obj, context, path=None):
        return -0.2087*obj + 159.045


TempYm = TempYmAdapter(BitsInteger(10))


class V3v3Adapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int(obj/4.0)

    def _decode(self, obj, context, path=None):
        return 4*obj


V3v3 = V3v3Adapter(BitsInteger(10))


class V5vAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int(obj/6.0)

    def _decode(self, obj, context, path=None):
        return 6*obj


V5v = V5vAdapter(BitsInteger(10))


# FC1
BOB = Struct(
    'sunsensor' / BitsInteger(10)[3],
    'paneltempX+' / TempXp,
    'paneltempX-' / TempXm,
    'paneltempY+' / TempYp,
    'paneltempY-' / TempYm,
    '3v3voltage' / V3v3,
    '3v3current' / BitsInteger(10),
    '5voltage' / V5v,
    )

# Nayif-1
ASIB = Struct(
    'sunsensor' / BitsInteger(10)[6],
    '3v3voltage' / V3v3,
    'imtquptime' / BitsInteger(20),
    '5voltage' / V5v,
    )


class RFTempAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int((obj-193.672)/-0.857)

    def _decode(self, obj, context, path=None):
        return -0.857*obj + 193.672


RFTemp = RFTempAdapter(Octet)


class RXCurrAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int(obj/0.636)

    def _decode(self, obj, context, path=None):
        return 0.636*obj


RXCurr = RXCurrAdapter(Octet)


class TXCurrAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int(obj/1.272)

    def _decode(self, obj, context, path=None):
        return 1.272*obj


TXCurr = TXCurrAdapter(Octet)


RF = Struct(
    'rxdoppler' / Octet,
    'rxrssi' / Octet,
    'temp' / RFTemp,
    'rxcurrent' / RXCurr,
    'tx3v3current' / RXCurr,
    'tx5vcurrent' / TXCurr,
    )


class PwrAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int((obj/5e-3)**(1.0/2.0629))

    def _decode(self, obj, context, path=None):
        return 5e-3*obj**2.0629


Pwr = PwrAdapter(Octet)


class PACurrAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int((obj-2.5435)/0.5496)

    def _decode(self, obj, context, path=None):
        return 0.5496*obj + 2.5435


PACurr = PACurrAdapter(Octet)

PA = Struct(
    'revpwr' / Pwr,
    'fwdpwr' / Pwr,
    'boardtemp' / Octet,  # TODO use lookup table
    'boardcurr' / PACurr,
    )

Ants = Struct(
    'temp' / Octet[2],  # TODO use ISIS manual
    'deployment' / Flag[4],
    )


# Also valid for Nayif-1
SWFC1 = Struct(
    'seqnumber' / BitsInteger(24),
    'dtmfcmdcount' / BitsInteger(6),
    'dtmflastcmd' / BitsInteger(5),
    'dtmfcmdsuccess' / Flag,
    'datavalid' / Flag[7],
    'eclipse' / Flag,
    'safemode' / Flag,
    'hwabf' / Flag,
    'swabf' / Flag,
    'deploymentwait' / Flag,
    )


RealTimeFC1 = BitStruct(
    'eps' / EPSFC1,
    'bob' / BOB,
    'rf' / RF,
    'pa' / PA,
    'ants' / Ants,
    'sw' / SWFC1,
    )


HighResolution = BitStruct(
    'sunsensor' / BitsInteger(10)[5],
    'photocurrent' / BitsInteger(15),
    'batteryvoltage' / BitsInteger(15),
    )


HRPayload = HighResolution[20]


class TempBlackChassisAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int((obj-75.244)/-0.024)

    def _decode(self, obj, context, path=None):
        return -0.024*obj + 75.244


TempBlackChassis = TempBlackChassisAdapter(BitsInteger(12))


class TempSilverChassisAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int((obj-74.750)/-0.024)

    def _decode(self, obj, context, path=None):
        return -0.024*obj + 74.750


TempSilverChassis = TempSilverChassisAdapter(BitsInteger(12))


class TempBlackPanelAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int((obj-75.039)/-0.024)

    def _decode(self, obj, context, path=None):
        return -0.024*obj + 75.039


TempBlackPanel = TempBlackPanelAdapter(BitsInteger(12))


class TempSilverPanelAdapter(Adapter):
    def _encode(self, obj, context, path=None):
        return int((obj-75.987)/-0.024)

    def _decode(self, obj, context, path=None):
        return -0.024*obj + 75.987


TempSilverPanel = TempSilverPanelAdapter(BitsInteger(12))


WholeOrbitFC1 = BitStruct(
    'tempblackchassis' / TempBlackChassis,
    'tempsilverchassis' / TempSilverChassis,
    'tempblackpanel' / TempBlackPanel,
    'tempsilverpanel' / TempSilverPanel,
    'paneltempX+' / TempXp,
    'paneltempX-' / TempXm,
    'paneltempY+' / TempYp,
    'paneltempY-' / TempYm,
    'photovoltage' / BitsInteger(16)[3],
    'photocurrent' / BitsInteger(16),
    'batteryvoltage' / BitsInteger(16),
    'systemcurrent' / BitsInteger(16),
    )

Callsign = Bytes(8)

FC2Battery = Struct(
    'direction' / Flag,
    'current' / Octet,
    'voltage' / Octet,
    'temp' / Octet,
    )

EPSFC2 = Struct(
    'sunlight' / Flag,
    'solarcurrent' / BitsInteger(10)[12],
    'solartemp' / Octet,
    'batteries' / FC2Battery[3],
    'batteryheater' / Flag,
    )

SWFC2 = Struct(
    'seqnumber' / BitsInteger(24),
    'dtmfcmdcount' / BitsInteger(6),
    'dtmflastcmd' / BitsInteger(5),
    'dtmfcmdsuccess' / Flag,
    )

RealTimeFC2 = BitStruct(
    'eps' / EPSFC2,
    'antstimeout' / Octet,
    'antsstatus' / BitsInteger(12),
    'antstemp' / Octet,
    'rf' / RF,
    'pa' / PA,
    'amacmode' / BitsInteger(3),
    'magnetometer' / BitsInteger(12)[3],
    'funtrxenable' / Flag,
    'funtrxsampleenable' / Flag,
    'modemanagermode' / BitsInteger(3),
    'modemanagercommsnominal' / Flag,
    'modemanagercommsstate' / BitsInteger(2),
    'tmtcmanageridleenable' / Flag,
    'tmtceventforwarding' / Flag,
    'tcbufferreceiveenable' / BitsInteger(3),
    'tcbuffersendenable' / BitsInteger(3),
    'obcsoftresetcount' / Octet,
    'epshardresetcount' / Octet,
    Padding(20),
    'sw' / SWFC2,
    )

FC2Battery0 = Struct(
    'current' / Octet,
    'voltage' / Octet,
    'temp' / Octet,
    )

FC2Battery2 = Struct(
    'direction' / Flag,
    'current' / Octet,
    'voltage' / Octet,
    )

WholeOrbitFC2 = BitStruct(
    'tempthermistor' / BitsInteger(12)[4],
    'solartemps' / BitsInteger(8)[5],
    'battery0' / FC2Battery0,
    'battery1' / FC2Battery,
    'battery2' / FC2Battery2,
    Padding(6),
    )

RealTimeNayif1 = BitStruct(
    'eps' / EPSNayif1,
    'imtq' / iMTQ,
    'asib' / ASIB,
    'rf' / RF,
    'pa' / PA,
    'ants' / Ants,
    'sw' / SWFC1,
    )

FitterMessage = Bytes(200)

Frame = Struct(
    'header' / Header,
    'extheader' / If(lambda c: c.header.satid == 'extended', Byte),
    'realtime' / Switch(lambda c: c.header.satid, {
        'FC1EM': RealTimeFC1,
        'FC1FM': RealTimeFC1,
        'FC2': RealTimeFC2,
        'extended': Switch(lambda c: c.extheader, {
            0x08: RealTimeNayif1,
            }, default=Bytes(54)),
        }),
    'payload' / If(lambda c: hasattr(c.header.frametype, '__getitem__'),
                   Switch(lambda c: c.header.frametype[:2], {
                    'WO': Bytes(200),
                    'HR': HRPayload,
                    'FM': FitterMessage,
                    })
                   ),
    )


def WholeOrbit(satid):
    if satid == 'FC1EM' or satid == 'FC1FM':
        return WholeOrbitFC1
    if satid == 'FC2':
        return WholeOrbitFC2
    return Bytes(23)


WHOLEORBIT_SIZE = 23
PAYLOAD_SIZE = 200
WHOLEORBIT_MAX = 12


class Funcube:
    """Telemetry parser for FUNcube

    This is a stateful parser that reassembles Whole Orbit data
    from consecutive frames
    """
    def __init__(self):
        self.last_chunk = None
        self.last_seq = None

    def parse(self, packet):
        if len(packet) != 256:
            return

        data = Frame.parse(packet)

        if not data:
            return

        out = list()

        out.append(f'Frame type {data.header.frametype}')
        if not hasattr(data.header.frametype, '__getitem__'):
            print('Unknown frame type. Not processing frame.')
            return

        out.append('-'*40)
        out.append('Realtime telemetry:')
        out.append('-'*40)
        out.append(str(data.realtime))
        out.append('-'*40)
        if data.header.frametype[:2] == 'FM':
            out.append(f'Fitter Message {data.header.frametype[2]}')
            out.append('-'*40)
            out.append(str(data.payload))
        elif data.header.frametype[:2] == 'HR':
            out.append(f'High resolution {data.header.frametype[2]}')
            out.append('-'*40)
            out.append(str(data.payload))
        elif data.header.frametype[:2] == 'WO':
            chunk = int(data.header.frametype[2:])
            try:
                seq = data.realtime.search('seqnumber')
            except AttributeError:
                out.append(
                    'Unknown realtime format. Unable to get seqnumber.\n')
                return '\n'.join(out)
            remaining = (PAYLOAD_SIZE*chunk) % WHOLEORBIT_SIZE
            recover = True
            if chunk != 0:
                if self.last_chunk == chunk - 1 and self.last_seq == seq:
                    # Can recover for last WO packet
                    wo = self.last_wo + data.payload[:-remaining]
                else:
                    recover = False
                    last_chunk_remaining = (
                        (PAYLOAD_SIZE*(chunk-1)) % WHOLEORBIT_SIZE)
                    wo = data.payload[
                        WHOLEORBIT_SIZE-last_chunk_remaining:-remaining]
            else:
                wo = data.payload[:-remaining]
            assert len(wo) % WHOLEORBIT_SIZE == 0
            wos = WholeOrbit(data.header.satid)[
                len(wo) // WHOLEORBIT_SIZE].parse(wo)
            self.last_chunk = chunk
            self.last_wo = data.payload[-remaining:]
            self.last_seq = seq
            out.append(f'Whole orbit {chunk}')
            if not recover:
                out.append('(could not recover data from previous beacon)')
            out.append('-'*40)
            out.append(str(wos))
            if chunk == WHOLEORBIT_MAX:
                out.append('-'*40)
                # Callsign included
                out.append(f'Callsign: {Callsign.parse(self.last_wo)}')
        out.append('')

        return '\n'.join(out)


funcube = Funcube()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 Stefano Speretta <s.speretta@tudelft.nl>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *
from .ax25 import Header
from ..adapters import LinearAdapter
from ..adapters import AffineAdapter
import math


partialBootCounter = 0
partialMeBoZpI = 0
partialMeBoZnI = 0
ComboI = 0
RAP1TXI = 0
RAP2RXI = 0
RAP2TXI = 0
RAP2FWDP = 0
RAP1REFP = 0
RAP2REFP = 0
RAP2RSSI = 0
RAP1Doppler = 0
RAP2Doppler = 0
RAP2T = 0


class ProcessFrameID(Adapter):
    def _decode(self, obj, context, path=None):
        global partialBootCounter
        partialBootCounter = (int(obj) >> 2) & 0x3F
        return int(obj) & 3


class ProcessBootCounter(Adapter):
    def _decode(self, obj, context, path=None):
        global partialBootCounter
        return ((int(obj) & 0x3F) << 6) + partialBootCounter


class ProcessFM430_I(Adapter):
    def _decode(self, obj, context, path=None):
        global partialMeBoZp
        partialMeBoZp = (int(obj) >> 2) & 0x3F
        return float((int(obj) >> 8) | ((int(obj) & 0x03) << 8)) * 0.395


class ProcessMeBoZp_I(Adapter):
    def _decode(self, obj, context, path=None):
        global partialMeBoZpI
        global partialMeBoZnI
        partialMeBoZnI = (int(obj) >> 4) & 0x0F
        return float(((int(obj) & 0x0F) << 6) + partialMeBoZp) * 0.395


class ProcessMeBoZn_I(Adapter):
    def _decode(self, obj, context, path=None):
        global partialMeBoZnI
        global ComboI
        ComboI = (int(obj) >> 6) & 0x03
        return float(((int(obj) & 0x3F) << 4) + partialMeBoZnI) * 0.395


class ProcessCombo_I(Adapter):
    def _decode(self, obj, context, path=None):
        global ComboI
        return float(((int(obj) & 0xFF) << 2) + ComboI) * 0.395


class ProcessRAP1Rx_I(Adapter):
    def _decode(self, obj, context, path=None):
        global RAP1TXI
        RAP1TXI = (int(obj) >> 2) & 0x3F
        return float((int(obj) >> 8) | ((int(obj) & 0x03) << 8)) * 0.395


class ProcessRAP1Tx_I(Adapter):
    def _decode(self, obj, context, path=None):
        global RAP1TXI
        global RAP2RXI
        RAP2RXI = (int(obj) >> 4) & 0x0F
        return float(((int(obj) & 0x0F) << 6) + RAP1TXI) * 0.395


class ProcessRAP2Rx_I(Adapter):
    def _decode(self, obj, context, path=None):
        global RAP2RXI
        global RAP2TXI
        RAP2TXI = (int(obj) >> 6) & 0x03
        return float(((int(obj) & 0x3F) << 4) + RAP2RXI) * 0.395


class ProcessRAP2Tx_I(Adapter):
    def _decode(self, obj, context, path=None):
        global RAP2TXI
        return float(((int(obj) & 0xFF) << 2) + RAP2TXI) * 0.395


class ProcessRAP1Fwd_P(Adapter):
    def _decode(self, obj, context, path=None):
        global RAP2FWDP
        RAP2FWDP = (int(obj) >> 2) & 0x3F
        return (
            float((int(obj) >> 8) | ((int(obj) & 0x03) << 8)) ** 2
            * 0.000478)


class ProcessRAP2Fwd_P(Adapter):
    def _decode(self, obj, context, path=None):
        global RAP2FWDP
        global RAP1REFP
        RAP1REFP = (int(obj) >> 4) & 0x0F
        return float(((int(obj) & 0x0F) << 6) + RAP2FWDP) ** 2 * 0.000478


class ProcessRAP1Ref_P(Adapter):
    def _decode(self, obj, context, path=None):
        global RAP1REFP
        global RAP2REFP
        RAP2REFP = (int(obj) >> 6) & 0x03
        return float(((int(obj) & 0x3F) << 4) + RAP1REFP) ** 2 * 0.000478


class ProcessRAP2Ref_P(Adapter):
    def _decode(self, obj, context, path=None):
        global RAP2REFP
        return float(((int(obj) & 0xFF) << 2) + RAP2REFP) ** 2 * 0.000478


class ProcessRAP1RSSI_P(Adapter):
    def _decode(self, obj, context, path=None):
        global RAP2RSSI
        RAP2RSSI = (int(obj) >> 2) & 0x3F
        return float((int(obj) >> 8) | ((int(obj) & 0x03) << 8)) * 3.23


class ProcessRAP2RSSI_P(Adapter):
    def _decode(self, obj, context, path=None):
        global RAP2RSSI
        global RAP1Doppler
        RAP1Doppler = (int(obj) >> 4) & 0x0F
        return float(((int(obj) & 0x0F) << 6) + RAP2RSSI) * 3.23


class ProcessRAP1Doppler_V(Adapter):
    def _decode(self, obj, context, path=None):
        global RAP1Doppler
        global RAP2Doppler
        RAP2Doppler = (int(obj) >> 6) & 0x03
        return float(((int(obj) & 0x3F) << 4) + RAP1Doppler) * 3.23


class ProcessRAP2Doppler_V(Adapter):
    def _decode(self, obj, context, path=None):
        global RAP2Doppler
        return float(((int(obj) & 0xFF) << 2) + RAP2Doppler) * 3.23


class ProcessRAP1_T(Adapter):
    def _decode(self, obj, context, path=None):
        global RAP2T
        RAP2T = (int(obj) >> 2) & 0x3F
        return float((int(obj) >> 8) | ((int(obj) & 0x03) << 8)) * 0.213 - 68.1


class ProcessRAP2_T(Adapter):
    def _decode(self, obj, context, path=None):
        global RAP2T
        return float(((int(obj) & 0x0F) << 6) + RAP2T) * 0.213 - 68.1


class BitArrayToByteArray(Adapter):
    def _decode(self, obj, context, path=None):
        sz = int(math.ceil(len(obj)/8))
        print("sie:  " + str(sz))
        out = bytearray(sz)
        for x in range(0, sz):
            for y in range(0, 7):
                out[sz - 1 - x] = out[sz - 1 - x] | (obj[x * 8 + y] << y)
        return out


OperationalMode = Enum(BitsInteger(4),
                       Idle=0,
                       Deployment=1,
                       Basic=2,
                       Science=3,
                       Transponder=4)

Housekeeping = BitStruct(
    'PICStatus_EMP' / BitsInteger(1),
    'RXcmd_ID' / BitsInteger(1),
    'SuccessfulBootCounter' / ProcessBootCounter(BitsInteger(6)),
    'PICStatus_CEP' / BitsInteger(1),
    'PICStatus_MDP1' / BitsInteger(1),
    'PICStatus_MDP2' / BitsInteger(1),
    'PICStatus_RBP1' / BitsInteger(1),
    'PICStatus_RBP2' / BitsInteger(1),
    'PICStatus_RCP1' / BitsInteger(1),
    'PICStatus_RCP2' / BitsInteger(1),
    'PICStatus_AWP' / BitsInteger(1),
    'PICStatus_ADP1' / BitsInteger(1),
    'PICStatus_ADP2' / BitsInteger(1),
    'PICStatus_ADP3' / BitsInteger(1),
    'PICStatus_ADP4' / BitsInteger(1),
    'PICStatus_MEP1' / BitsInteger(1),
    'PICStatus_MEP2' / BitsInteger(1),
    'PICStatus_REP1' / BitsInteger(1),
    'PICStatus_REP2' / BitsInteger(1),
    'MemoryStatus1' / BitArrayToByteArray(Bytes(64)),
    'MemoryStatus2' / BitArrayToByteArray(Bytes(64)),
    'deployStatusVector_SP_ZpXm' / BitsInteger(1),
    'deployStatusVector_SP_ZpXp' / BitsInteger(1),
    'deployStatusVector_MAB_ZmYp' / BitsInteger(1),
    'deployStatusVector_MAB_ZmXp' / BitsInteger(1),
    'deployStatusVector_MAB_ZmYm' / BitsInteger(1),
    'deployStatusVector_MAB_ZmXm' / BitsInteger(1),
    'deployStatusVector_SP_ZmYm' / BitsInteger(1),
    'deployStatusVector_SP_ZmYp' / BitsInteger(1),
    'operationalMode' / OperationalMode,
    'deployStatusVector_MAB_ZpXm' / BitsInteger(1),
    'deployStatusVector_MAB_ZpYp' / BitsInteger(1),
    'deployStatusVector_MAB_ZpXp' / BitsInteger(1),
    'deployStatusVector_MAB_ZpYm' / BitsInteger(1),
    'bus_V_sys' / LinearAdapter(1/0.049, BitsInteger(8)),  # V
    'bus_V_dep' / LinearAdapter(1/0.049, BitsInteger(8)),  # V
    'OBC_T' / AffineAdapter(1/0.6875, 80, BitsInteger(8)),  # degC
    'SP_I_ZpXp' / LinearAdapter(1/1.95, BitsInteger(8)),  # mA
    'SP_I_ZmYm' / LinearAdapter(1/1.95, BitsInteger(8)),  # mA
    'SP_I_ZpXm' / LinearAdapter(1/1.95, BitsInteger(8)),  # mA
    'SP_I_ZmYp' / LinearAdapter(1/1.95, BitsInteger(8)),  # mA
    'FM430_I' / ProcessFM430_I(BitsInteger(16)),  # mA
    'MeBo_Zp_I' / ProcessMeBoZp_I(BitsInteger(8)),  # mA
    'MeBo_Zn_I' / ProcessMeBoZn_I(BitsInteger(8)),  # mA
    'Combo_I' / ProcessCombo_I(BitsInteger(8)),  # mA
    'RAP1Rx_I' / ProcessRAP1Rx_I(BitsInteger(16)),  # mA
    'RAP1Tx_I' / ProcessRAP1Tx_I(BitsInteger(8)),  # mA
    'RAP2Rx_I' / ProcessRAP2Rx_I(BitsInteger(8)),  # mA
    'RAP2Tx_I' / ProcessRAP2Tx_I(BitsInteger(8)),  # mA
    'RAP1FwdPower' / ProcessRAP1Fwd_P(BitsInteger(16)),  # mW
    'RAP2FwdPower' / ProcessRAP2Fwd_P(BitsInteger(8)),  # mW
    'RAP1ReflPower' / ProcessRAP1Ref_P(BitsInteger(8)),  # mW
    'RAP2ReflPower' / ProcessRAP2Ref_P(BitsInteger(8)),  # mW
    'RAP1RSSI' / ProcessRAP1RSSI_P(BitsInteger(16)),  # mV
    'RAP2RSSI' / ProcessRAP2RSSI_P(BitsInteger(8)),  # mV
    'RAP1Doppler_V' / ProcessRAP1Doppler_V(BitsInteger(8)),  # V
    'RAP2Doppler_V' / ProcessRAP2Doppler_V(BitsInteger(8)),  # V
    'RAP1T' / ProcessRAP1_T(BitsInteger(16)),  # degC
    'RAP2T' / ProcessRAP2_T(BitsInteger(8)),  # degC
    'AWSS1' / BitArrayToByteArray(Bytes(168)),
    'AWSS2' / BitArrayToByteArray(Bytes(168)),

    # V
    'ReferenceDiode_Voltage_ZpXp' / LinearAdapter(1/0.002, BitsInteger(8)),
    # V
    'ReferenceDiode_Voltage_ZpXn' / LinearAdapter(1/0.002, BitsInteger(8)),
    # mA
    'ReferenceDiode_Current_ZpXp' / LinearAdapter(1/0.004, BitsInteger(8)),
    # mA
    'ReferenceDiode_Current_ZpXn' / LinearAdapter(1/0.004, BitsInteger(8)),
    # V
    'ReferenceDiode_Voltage_ZnYm' / LinearAdapter(1/0.002, BitsInteger(8)),
    # V
    'ReferenceDiode_Voltage_ZnYp' / LinearAdapter(1/0.002, BitsInteger(8)),
    # mA
    'ReferenceDiode_Current_ZnYm' / LinearAdapter(1/0.004, BitsInteger(8)),
    # mA
    'ReferenceDiode_Current_ZnYp' / LinearAdapter(1/0.004, BitsInteger(8))
    )

Payload = BitStruct(
    'TFSC_Current_ZpXp_0' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZpXp_0' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZpXp_1' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZpXp_1' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZpXp_2' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZpXp_2' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZpXp_3' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZpXp_3' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZpXp_4' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZpXp_4' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZpXp_5' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZpXp_5' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZpXp_6' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZpXp_6' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZpXp_7' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZpXp_7' / LinearAdapter(1/0.008, BitsInteger(8)),  # V

    'TFSC_Current_ZpXm_0' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZpXm_0' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZpXm_1' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZpXm_1' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZpXm_2' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZpXm_2' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZpXm_3' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZpXm_3' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZpXm_4' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZpXm_4' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZpXm_5' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZpXm_5' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZpXm_6' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZpXm_6' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZpXm_7' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZpXm_7' / LinearAdapter(1/0.008, BitsInteger(8)),  # V

    'TFSC_Current_ZmXp_0' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZmXp_0' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZmXp_1' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZmXp_1' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZmXp_2' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZmXp_2' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZmXp_3' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZmXp_3' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZmXp_4' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZmXp_4' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZmXp_5' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZmXp_5' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZmXp_6' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZmXp_6' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZmXp_7' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZmXp_7' / LinearAdapter(1/0.008, BitsInteger(8)),  # V

    'TFSC_Current_ZmXm_0' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZmXm_0' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZmXm_1' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZmXm_1' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZmXm_2' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZmXm_2' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZmXm_3' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZmXm_3' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZmXm_4' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZmXm_4' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZmXm_5' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZmXm_5' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZmXm_6' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZmXm_6' / LinearAdapter(1/0.008, BitsInteger(8)),  # V
    'TFSC_Current_ZmXm_7' / LinearAdapter(1/0.004, BitsInteger(8)),  # A
    'TFSC_Voltage_ZmXm_7' / LinearAdapter(1/0.008, BitsInteger(8)),  # V

    # degC
    'TFSC_Temperature_ZpXp' / AffineAdapter(1/1.492, -235.0, BitsInteger(8)),
    # degC
    'TFSC_Temperature_ZpXm' / AffineAdapter(1/1.389, -235.0, BitsInteger(8)),

    # degC
    'TFSC_Temperature_ZmXp' / AffineAdapter(1/1.445, -235.0, BitsInteger(8)),
    # degC
    'TFSC_Temperature_ZmXm' / AffineAdapter(1/1.345, -235.0, BitsInteger(8)),

    # V
    'ReferenceDiode_Voltage_ZpXp' / LinearAdapter(1/0.002, BitsInteger(8)),
    # V
    'ReferenceDiode_Voltage_ZpXn' / LinearAdapter(1/0.002, BitsInteger(8)),
    # mA
    'ReferenceDiode_Current_ZpXp' / LinearAdapter(1/0.004, BitsInteger(8)),
    # mA
    'ReferenceDiode_Current_ZpXn' / LinearAdapter(1/0.004, BitsInteger(8)),
    # V
    'ReferenceDiode_Voltage_ZnYm' / LinearAdapter(1/0.002, BitsInteger(8)),
    # V
    'ReferenceDiode_Voltage_ZnYp' / LinearAdapter(1/0.002, BitsInteger(8)),
    # mA
    'ReferenceDiode_Current_ZnYm' / LinearAdapter(1/0.004, BitsInteger(8)),
    # mA
    'ReferenceDiode_Current_ZnYp' / LinearAdapter(1/0.004, BitsInteger(8)),

    'AWSS' / BitArrayToByteArray(Bytes(168))
    )

Auxiliary = Struct(
    Padding(6)
    )

delfic3 = Struct(
    'ax25_header' / Header,
    'BootNumber' / BytesInteger(2, swapped=True),
    'FrameNumber' / BytesInteger(2, swapped=True),
    'FrameID' / ProcessFrameID(BytesInteger(1)),
    'packet' / Switch(lambda c: (c.FrameID), {
        1: Payload,
        2: Housekeeping})
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 jgromes <gromes.jan@gmail.com>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import struct

from gnuradio import gr
import numpy
import pmt


crc_table = [
    0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50A5, 0x60C6, 0x70E7,
    0x8108, 0x9129, 0xA14A, 0xB16B, 0xC18C, 0xD1AD, 0xE1CE, 0xF1EF,
    0x1231, 0x0210, 0x3273, 0x2252, 0x52B5, 0x4294, 0x72F7, 0x62D6,
    0x9339, 0x8318, 0xB37B, 0xA35A, 0xD3BD, 0xC39C, 0xF3FF, 0xE3DE,
    0x2462, 0x3443, 0x0420, 0x1401, 0x64E6, 0x74C7, 0x44A4, 0x5485,
    0xA56A, 0xB54B, 0x8528, 0x9509, 0xE5EE, 0xF5CF, 0xC5AC, 0xD58D,
    0x3653, 0x2672, 0x1611, 0x0630, 0x76D7, 0x66F6, 0x5695, 0x46B4,
    0xB75B, 0xA77A, 0x9719, 0x8738, 0xF7DF, 0xE7FE, 0xD79D, 0xC7BC,
    0x48C4, 0x58E5, 0x6886, 0x78A7, 0x0840, 0x1861, 0x2802, 0x3823,
    0xC9CC, 0xD9ED, 0xE98E, 0xF9AF, 0x8948, 0x9969, 0xA90A, 0xB92B,
    0x5AF5, 0x4AD4, 0x7AB7, 0x6A96, 0x1A71, 0x0A50, 0x3A33, 0x2A12,
    0xDBFD, 0xCBDC, 0xFBBF, 0xEB9E, 0x9B79, 0x8B58, 0xBB3B, 0xAB1A,
    0x6CA6, 0x7C87, 0x4CE4, 0x5CC5, 0x2C22, 0x3C03, 0x0C60, 0x1C41,
    0xEDAE, 0xFD8F, 0xCDEC, 0xDDCD, 0xAD2A, 0xBD0B, 0x8D68, 0x9D49,
    0x7E97, 0x6EB6, 0x5ED5, 0x4EF4, 0x3E13, 0x2E32, 0x1E51, 0x0E70,
    0xFF9F, 0xEFBE, 0xDFDD, 0xCFFC, 0xBF1B, 0xAF3A, 0x9F59, 0x8F78,
    0x9188, 0x81A9, 0xB1CA, 0xA1EB, 0xD10C, 0xC12D, 0xF14E, 0xE16F,
    0x1080, 0x00A1, 0x30C2, 0x20E3, 0x5004, 0x4025, 0x7046, 0x6067,
    0x83B9, 0x9398, 0xA3FB, 0xB3DA, 0xC33D, 0xD31C, 0xE37F, 0xF35E,
    0x02B1, 0x1290, 0x22F3, 0x32D2, 0x4235, 0x5214, 0x6277, 0x7256,
    0xB5EA, 0xA5CB, 0x95A8, 0x8589, 0xF56E, 0xE54F, 0xD52C, 0xC50D,
    0x34E2, 0x24C3, 0x14A0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
    0xA7DB, 0xB7FA, 0x8799, 0x97B8, 0xE75F, 0xF77E, 0xC71D, 0xD73C,
    0x26D3, 0x36F2, 0x0691, 0x16B0, 0x6657, 0x7676, 0x4615, 0x5634,
    0xD94C, 0xC96D, 0xF90E, 0xE92F, 0x99C8, 0x89E9, 0xB98A, 0xA9AB,
    0x5844, 0x4865, 0x7806, 0x6827, 0x18C0, 0x08E1, 0x3882, 0x28A3,
    0xCB7D, 0xDB5C, 0xEB3F, 0xFB1E, 0x8BF9, 0x9BD8, 0xABBB, 0xBB9A,
    0x4A75, 0x5A54, 0x6A37, 0x7A16, 0x0AF1, 0x1AD0, 0x2AB3, 0x3A92,
    0xFD2E, 0xED0F, 0xDD6C, 0xCD4D, 0xBDAA, 0xAD8B, 0x9DE8, 0x8DC9,
    0x7C26, 0x6C07, 0x5C64, 0x4C45, 0x3CA2, 0x2C83, 0x1CE0, 0x0CC1,
    0xEF1F, 0xFF3E, 0xCF5D, 0xDF7C, 0xAF9B, 0xBFBA, 0x8FD9, 0x9FF8,
    0x6E17, 0x7E36, 0x4E55, 0x5E74, 0x2E93, 0x3EB2, 0x0ED1, 0x1EF0]


def crc16(data, initial, final):
    crc = initial

    for byte in data:
        tbl_idx = ((crc >> 8) ^ byte) & 0xff
        crc = (crc_table[tbl_idx] ^ (crc << 8)) & 0xffff

    return (crc & 0xffff) ^ final


class sx12xx_check_crc(gr.basic_block):
    """Check SX12xx FSK packet CRC-16

    The CRC-16 is calculated from length field and payload.
    """
    def __init__(self, verbose, initial, final):
        gr.basic_block.__init__(
            self,
            name='sx12xx_check_crc',
            in_sig=[],
            out_sig=[])

        self.verbose = verbose
        self.initial = initial
        self.final = final

        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('ok'))
        self.message_port_register_out(pmt.intern('fail'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        packet_orig = pmt.u8vector_elements(msg)
        packet = bytes(packet_orig)

        if len(packet) < 3:
            return

        packet_out = packet_orig[:-2]
        packet_crc = int.from_bytes(packet[-2:], 'big')
        msg_out = pmt.cons(
            pmt.car(msg_pmt),
            pmt.init_u8vector(len(packet_out), packet_out))
        res = crc16(packet[:-2], self.initial, self.final)
        if res == packet_crc:
            if self.verbose:
                print('CRC OK')
            self.message_port_pub(pmt.intern('ok'), msg_out)
        else:
            if self.verbose:
                print('CRC failed')
            self.message_port_pub(pmt.intern('fail'), msg_out)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

'''
gr-satellites demodulators components

The deframers transform IQ or real samples into soft symbols. Only
demodulation is done, with no packet boundary detection.

The input to these hierarchical blocks is a stream of samples
and the output is a stream of soft symbols.
'''

from .afsk_demodulator import afsk_demodulator
from .bpsk_demodulator import bpsk_demodulator
from .fsk_demodulator import fsk_demodulator



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 jgromes <gromes.jan@gmail.com>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *

from ..adapters import LinearAdapter
from .fossasat import *

fossasat_2 = Struct(
    'callsign' / Const(b"FOSSASAT-2"),
    'func_id' / FuncId,
    'opt_data_len' / Optional(Int8ul),
    'payload' / Switch(this.func_id, {
        'RESP_PONG': Pong,
        'RESP_REPEATED_MESSAGE': RepeatedMessage,
        'RESP_REPEATED_MESSAGE_CUSTOM': RepeatedMessage,
        'RESP_SYSTEM_INFO': SystemInfo,
        'RESP_PACKET_INFO': PacketInfo,
        'RESP_STATISTICS': Statistics,
        'RESP_FULL_SYSTEM_INFO': FullSystemInfo,
        'RESP_STORE_AND_FORWARD_ASSIGNED_SLOT': StoreAndForwardAssigned,
        'RESP_FORWARDED_MESSAGE': ForwardedMessage,
        'RESP_PUBLIC_PICTURE': CameraPicture,
        'RESP_DEPLOYMENT_STATE': DeploymentState,
        'RESP_CAMERA_STATE': CameraState,
        'RESP_RECORDED_IMU': RecordedIMU,
        'RESP_MANUAL_ACS_RESULT': ManualACSResult,
        'RESP_GPS_LOG': GPSLog,
        'RESP_GPS_LOG_STATE': GPSLogState,
        'RESP_FLASH_CONTENTS': FlashContents,
        'RESP_CAMERA_PICTURE': CameraPicture,
        'RESP_CAMERA_PICTURE_LENGTH': CameraPictureLength,
        'RESP_GPS_COMMAND_RESPONSE': GPSCommandResponse,
        'RESP_ACKNOWLEDGE': Acknowledge,
        }, default=HexDump(GreedyBytes)),
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

import struct

from gnuradio import gr
import numpy
import pmt


class check_address(gr.basic_block):
    """docstring for block check_address"""
    def __init__(self, address, direction):
        gr.basic_block.__init__(
            self,
            name='check_address',
            in_sig=[],
            out_sig=[])

        a = address.split('-')
        self.callsign = a[0]
        self.ssid = int(a[1]) if len(a) > 1 else None
        self.direction = direction

        self.message_port_register_in(pmt.intern('in'))
        self.set_msg_handler(pmt.intern('in'), self.handle_msg)
        self.message_port_register_out(pmt.intern('ok'))
        self.message_port_register_out(pmt.intern('fail'))

    def handle_msg(self, msg_pmt):
        msg = pmt.cdr(msg_pmt)
        if not pmt.is_u8vector(msg):
            print('[ERROR] Received invalid message type. Expected u8vector')
            return
        packet = bytes(pmt.u8vector_elements(msg))

        # Check packet length.
        # An AX.25 header with 2 addresses, control and PID is 16 bytes.
        if len(packet) < 16:
            self.message_port_pub(pmt.intern('fail'), msg_pmt)
            return

        if self.direction == 'to':
            address = packet[:7]
        else:
            address = packet[7:14]

        callsign = [c >> 1 for c in address[:6]]
        callsign = bytes(callsign).decode('ascii').rstrip(' ')
        ssid = (address[6] >> 1) & 0x0f

        if (callsign != self.callsign
                or (self.ssid is not None and ssid != self.ssid)):
            # Incorrect address
            self.message_port_pub(pmt.intern('fail'), msg_pmt)
        else:
            # Correct address
            self.message_port_pub(pmt.intern('ok'), msg_pmt)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2018-2019 Daniel Estevez <daniel@destevez.net>
#
# This file is part of gr-satellites
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from construct import *

from ..adapters import UNIXTimestampAdapter
from .csp import CSPHeader


Timestamp = UNIXTimestampAdapter(Int32ub)

Beacon0EPS = Struct(
    'timestamp' / Timestamp,
    'pv_v' / Int16ub[3],
    'batt_v' / Int16ub,
    'output_cur' / Int16ub[7],
    'pv_cur' / Int16ub[3],
    'batt_in_cur' / Int16ub,
    'batt_out_cur' / Int16ub,
    'temp' / Int16ub[6],
    'batt_mode' / Int8ub
    )

Beacon0COM = Struct(
    'timestamp' / Timestamp,
    'temp' / Int16sb[2],
    'rssi' / Int16sb,
    'rferr' / Int16sb,
    'rssi_bgnd' / Int16sb
    )

Beacon0OBC = Struct(
    'timestamp' / Timestamp,
    'cur' / Int16ub[6],
    'temp' / Int16sb[2]
    )

Beacon0 = Struct(
    'beacon_type' / Const(b'\x00'),
    'eps' / Beacon0EPS,
    'com' / Beacon0COM,
    'obc' / Beacon0OBC
    )

Beacon1EPS = Struct(
    'timestamp' / Timestamp,
    'wdt_i2c' / Int32ub,
    'wdt_gnd' / Int32ub,
    'boot_count' / Int32ub,
    'wdt_i2c_count' / Int32ub,
    'wdt_gnd_count' / Int32ub,
    'wdt_csp_count' / Int32ub[2],
    'wdt_csp' / Int8ub[2],
    'boot_cause' / Int8ub,
    'latchup' / Int16ub[6],
    'out_val' / Int8ub[8],
    'ppt_mode' / Int8ub
    )

Beacon1COM = Struct(
    'timestamp' / Timestamp,
    'tx_duty' / Int8ub,
    'total_tx_count' / Int32ub,
    'total_rx_count' / Int32ub,
    'total_tx_bytes' / Int32ub,
    'total_rx_bytes' / Int32ub,
    'boot_count' / Int16ub,
    'boot_cause' / Int32ub,
    'tx_bytes' / Int32ub,
    'rx_bytes' / Int32ub,
    'config' / Int8ub,
    'tx_count' / Int32ub,
    'rx_count' / Int32ub
    )

Beacon1OBC = Struct(
    'timestamp' / Timestamp,
    'pwr' / Int8ub[6],
    'sw_count' / Int16ub,
    'filesystem' / Int8ub,
    'boot_count' / Int16ub,
    'boot_cause' / Int32ub,
    'clock' / Timestamp
    )

Beacon1 = Struct(
    'beacon_type' / Const(b'\x01'),
    'eps' / Beacon1EPS,
    'com' / Beacon1COM,
    'obc' / Beacon1OBC
    )

suomi100 = Struct(
    'header' / CSPHeader,
    'payload' / Select(Beacon0, Beacon1)
    )
